Imports System.Windows.Forms
Imports System.Threading

'UPDATE LOG
'==========
'****************************************************************************
'VER    DATE      CHANGE
'3.0    31/07/04  UPDATE FOR NEW RPIC MODULE - 80 Bytes IO and 19200bps Comms
'****************************************************************************

Class Hardware
    'Handles All Interaction with RPC Hardware using RS232 Link
    'Refer to MERG Interface Specification G16/4 Issue 2
    'Sets up and maintains array of current Input and Output bits
    'Synchronises Hardware with Array

    Inherits Object

    Private Max_Input_Output_Bytes As Integer = 80 'From MERG Spec G16/4 - Modified 32-80 by new SW. Only Type 0 Messages now allowed.
    Private UpdateInterval As Integer = 50 'milliseconds between Hardware Synchronization.
    Friend HardwareSystem As String = "RPC Hardware"
    Friend CommsType As String = "RS232"
    Private RPC_Out As Boolean()
    Private RPC_In As Boolean()
    Private ConnectionInProgress As Boolean = False
    Private CommsFailCount As Integer = 0
    Private oCP As New Rs232()
    Private intCommPort, intBaud, intData As Integer
    Private bytStop As Rs232.DataStopBit
    Private bytParity As Rs232.DataParity
    Private MaxRPCBit As Integer
    Private mConnected As Boolean = False

    Private HW_Sync As New SyncHW()
    Private Sync As New Thread(AddressOf HW_Sync.Go)

    Public Sub New(ByVal Max_RPC_Bytes As Integer)
        Data.MaxRPCBytes = Max_RPC_Bytes
        Sync.Name = "Sync RPC"
        Sync.Start()
    End Sub

    Public ReadOnly Property Connected() As Boolean
        Get
            Return mConnected
        End Get
    End Property

    Public ReadOnly Property Max_RPCBit() As Integer
        Get
            Return MaxRPCBit
        End Get
    End Property

    Public ReadOnly Property RPC_Inputs() As Boolean()
        Get
            Return RPC_In
        End Get

    End Property

    Public ReadOnly Property RPC_Outputs() As Boolean()
        Get
            Return RPC_Out
        End Get

    End Property

    Public ReadOnly Property SyncInterval() As Integer
        Get
            Return UpdateInterval
        End Get
    End Property

    Public ReadOnly Property ConnectionInUse() As Boolean
        Get
            Return ConnectionInProgress
        End Get
    End Property
    Public Sub SetupHardware(ByVal Max_RPC_Bytes As Integer)
        Dim i As Integer
        If Max_RPC_Bytes = -1 Then
            Dim RPCConfig As New frmRPCConfiguration()
            Max_RPC_Bytes = RPCConfig.Get_RPC_Config(Max_Input_Output_Bytes)
            If Max_RPC_Bytes = Data.MaxRPCBytes Then
                'No Change
            Else
                If Max_RPC_Bytes * 8 - 1 < Data.GetHighestAddress() Then
                    MessageBox.Show("Invalid - You can not reduce the ammount of I/O while Devices reference bits to be removed. Changes ignored.", mrcc)
                    Max_RPC_Bytes = Data.MaxRPCBytes
                Else
                    Data.MaxRPCBytes = Max_RPC_Bytes
                    MessageBox.Show("RPC Setup Configured - Bits from 0 to " & Max_RPC_Bytes * 8 - 1 & " now available.", mrcc)
                End If
            End If
            RPCConfig.Close()
            RPCConfig = Nothing
        End If
        If Max_RPC_Bytes <> -1 Then

            MaxRPCBit = 8 * Max_RPC_Bytes - 1
            Data.MaxRPCBytes = Max_RPC_Bytes
            RPC_Out = New Boolean(MaxRPCBit) {}
            RPC_In = New Boolean(MaxRPCBit) {}

            For i = 0 To MaxRPCBit
                RPC_Out(i) = False
                RPC_In(i) = False
            Next
        End If
    End Sub

    Public Sub SyncHardware()
        Type0Message(Data.MaxRPCBytes)
    End Sub

    Public Sub SimulateInput(ByVal RPCBit As Integer, ByVal State As Boolean)
        If RPCBit >= 0 And RPCBit <= MaxRPCBit Then
            RPC_In(RPCBit) = State
        End If
    End Sub

    Private Sub Read_Hardware(ByVal RPC_Bit As Integer) 'Reads Bit from Hardware to Array
        ' G16/4 Type 5 Message
        If RPC_Bit >= 0 And RPC_Bit <= MaxRPCBit Then
            Dim message As Byte() = New Byte(5) {}
            Dim i As Integer
            Dim inputmessage As Byte() = New Byte(1) {0, 0}
            Dim databytes As Byte() = New Byte(1) {0, 0}
            message(0) = "&HAA"
            message(1) = "&H55"
            message(2) = "&H05"
            message(3) = "&H02"
            message(4) = CByte(RPC_Bit)
            databytes(0) = message(4)
            message(5) = Calc_Checksum(databytes)
            If mConnected = True Then
                Try
                    oCP.Write(message)
                Catch ex As Exception
                    MessageBox.Show("Unable to Write to Comm Port", mrcc)
                End Try
            End If
            'Read Response
            Try
                oCP.Read(2)
            Catch ex As Exception
                MessageBox.Show("Unable to Read from Comm Port", mrcc)
            End Try
            inputmessage = oCP.InputStream()
            databytes(0) = inputmessage(0)
            If inputmessage(1) = Calc_Checksum(databytes) Then
                If inputmessage(0) = 1 Then
                    RPC_In(RPC_Bit) = True
                Else
                    RPC_In(RPC_Bit) = False
                End If
            End If
        End If
    End Sub

    Private Sub Write_Hardware(ByVal RPC_Bit As Integer) 'Writes Bit from Array to Hardware
        ' G16/4 Type 3(Set) or 4(Clear) Message
        If RPC_Bit >= 0 And RPC_Bit <= MaxRPCBit Then
            Dim message As Byte() = New Byte(5) {}
            Dim databytes As Byte() = New Byte(1) {0, 0}
            Dim i As Integer

            message(0) = "&HAA"
            message(1) = "&H55"
            message(3) = "&H02"
            If RPC_Out(RPC_Bit) = True Then
                'Type 3 Message - Set Bit
                message(2) = "&H03"
            Else
                'Type 4 Message - Clear Bit
                message(2) = "&H04"
            End If
            message(4) = CByte(RPC_Bit)
            databytes(0) = message(4)
            message(5) = Calc_Checksum(databytes)
            If mConnected = True Then
                Try
                    oCP.Write(message)
                Catch ex As Exception
                    MessageBox.Show("Unable to Write to Comm Port", mrcc)
                End Try
            End If
        End If
    End Sub

    Private Sub Type0Message(ByVal MaxRPCBytes As Integer) 'All Bytes Out, All Bytes Read
        ' G16/4 Type 0 Message
        Dim i As Integer, a As Integer
        Dim message As Byte() = New Byte(MaxRPCBytes + 4) {}
        Dim inputmessage As Byte() = New Byte(MaxRPCBytes) {}
        Dim databytes As Byte() = New Byte(MaxRPCBytes - 1) {}
        message(0) = "&HAA"
        message(1) = "&H55"
        message(2) = "&H00"
        message(3) = CByte(MaxRPCBytes + 1)
        For i = 0 To MaxRPCBytes - 1
            message(i + 4) = BitArray_Byte(i * 8)
            databytes(i) = message(i + 4)
        Next i
        message(MaxRPCBytes + 4) = Calc_Checksum(databytes)
        If mConnected = True And CommsFailCount <> 5 Then
            Try
                oCP.Write(message)
            Catch ex As Exception
                MessageBox.Show("Unable to Write to Comm Port", mrcc)
            End Try
            Sync.Sleep(1)
            'Read Response to Array
            Try
                oCP.Read(MaxRPCBytes + 1)
                inputmessage = oCP.InputStream()
                For i = 0 To MaxRPCBytes - 1
                    databytes(i) = inputmessage(i)
                Next i
                If inputmessage(MaxRPCBytes) = Calc_Checksum(databytes) Then
                    For i = 0 To MaxRPCBytes - 1
                        Byte_BitArray(databytes(i), i * 8)
                    Next i
                End If
                'Reset Failure Count if subsequent Read OK
                If CommsFailCount > 0 Then CommsFailCount = 0
            Catch ex As Exception
                'Wait 150ms for RPIC to Clear Buffers
                Sync.Sleep(150)
                CommsFailCount += 1
                If CommsFailCount = 5 Then MessageBox.Show("Unable to Read from Port - Change to Design Mode and Reconnect", mrcc)
            End Try
        End If
    End Sub

    Private Function BitArray_Byte(ByVal lowbit As Integer) As Byte
        Dim i As Integer
        Dim temp As Integer = 0
        For i = 0 To 7
            If RPC_Out(lowbit + i) = True Then temp += 2 ^ i
        Next
        Return CByte(temp)
    End Function

    Private Sub Byte_BitArray(ByVal inputbyte As Byte, ByVal lowbit As Integer)
        Dim i As Integer
        Dim temp As Integer = CInt(inputbyte)
        For i = 7 To 0 Step -1
            If temp > (2 ^ i) - 1 Then
                RPC_In(lowbit + i) = True
                temp -= (2 ^ i)
            Else
                RPC_In(lowbit + i) = False
            End If
        Next i
    End Sub

    Private Function Calc_Checksum(ByVal data As Byte()) As Byte
        Dim i As Integer = 0, chksum As Byte = 0
        For i = 0 To data.GetUpperBound(0)
            chksum = data(i) Xor chksum
        Next
        Return chksum
    End Function

    Public Function ReadBit(ByVal RPC_Bit As Integer) As Integer 'Returns Value from Array
        If RPC_Bit >= 0 And RPC_Bit <= MaxRPCBit Then
            If RPC_In(RPC_Bit) Then
                Return 1
            Else
                Return 0
            End If
        Else
            Return -1
        End If
    End Function 'Read

    Public Sub SetBit(ByVal RPC_Bit As Integer) 'Sets Output Bit in Array
        If RPC_Bit >= 0 And RPC_Bit <= MaxRPCBit Then
            RPC_Out(RPC_Bit) = True
        End If
    End Sub 'Set Bit

    Public Sub ClearBit(ByVal RPC_Bit As Integer) 'Clears Output Bit in Array
        If RPC_Bit >= 0 And RPC_Bit <= MaxRPCBit Then
            RPC_Out(RPC_Bit) = False
        End If
    End Sub 'ClearBit

    Public Function Connect_To_Hardware() As Boolean
        Dim fComSel As New ComPortSelect()
        CheckAvailablePorts()
        fComSel.ShowDialog()
        If fComSel.bGo Then
            intCommPort = CInt(Mid(fComSel.ComPortDropDown.SelectedItem, 4))
            intBaud = CInt(fComSel.BaudRate.SelectedItem)
            intData = 8
            bytParity = Rs232.DataParity.Parity_None
            bytStop = Rs232.DataStopBit.StopBit_1
            ConnectionInProgress = True
            oCP.Open(intCommPort, intBaud, intData, bytParity, bytStop, 4096)
            MessageBox.Show(HardwareSystem & " Connected via " & CommsType & " at " & intBaud & "bps on COM" & intCommPort, mrcc)
            fComSel.Close()
            fComSel = Nothing
            oCP.Timeout = 2000
            CommsFailCount = 0
            mConnected = True
            ConnectionInProgress = False
            Return True
        Else
            fComSel.Close()
            fComSel = Nothing
            Return False
        End If
    End Function 'Connect_to_ Hardware

    Public Sub Disconnect_Hardware()
        ConnectionInProgress = True
        Do Until HW_Sync.SyncInProgress = False
        Loop
        oCP.Close()
        mConnected = False
        ConnectionInProgress = False
        MessageBox.Show("Hardware can now be Disconnected", mrcc)
    End Sub 'Disconnect_Hardware

    Private Function IsPortAvailable(ByVal ComPort As Integer) As Boolean
        If HW.Connected = False And ConnectionInProgress = False Then
            Try
                oCP.Open(ComPort, 115200, 8, Rs232.DataParity.Parity_None, _
                    Rs232.DataStopBit.StopBit_1, 4096)
                ' If it makes it to here, then the Comm Port is available.
                oCP.Close()
                Return True
            Catch
                ' If it gets here, then the attempt to open the Comm Port
                '   was unsuccessful.
                Return False
            End Try
        End If
    End Function

    Private Sub CheckAvailablePorts()
        Dim i As Integer

        For i = 0 To 3
            mComs(i) = IsPortAvailable(i + 1)
        Next
    End Sub

    Public Sub EndSyncThread()
        HW_Sync.StopSync = True
    End Sub
End Class

Class SyncHW
    Dim stopme As Boolean = False
    Dim SyncNow As Boolean = False

    Sub Go()
        While stopme = False
            If HW.Connected And HW.ConnectionInUse = False Then
                SyncNow = True
                HW.SyncHardware()
                SyncNow = False
            End If
            Thread.Sleep(HW.SyncInterval)
        End While
    End Sub

    Public Property StopSync()
        Get
            Return stopme
        End Get
        Set(ByVal Value)
            stopme = Value
        End Set
    End Property

    Public ReadOnly Property SyncInProgress() As Boolean
        Get
            Return SyncNow
        End Get
    End Property

End Class
