Imports System.Collections

Module DisplayLimits
    'Grid Size
    Public Const TileSize = 16
    Public Const Max_SymbolCode = 159
    Public Const Max_X = 80
    Public Const Max_Y = 64
    Public CurrentX As Integer = 0
    Public CurrentY As Integer = 0
    Public SymbolStore As Bitmap() = New Bitmap(Max_SymbolCode) {}
    Public SymbolStoreSelect As Bitmap() = New Bitmap(Max_SymbolCode) {}

    Public Function CreateSymbolStore()
        Dim SymbolFile As Bitmap = Bitmap.FromFile("Symbols.png")
        Dim i As Integer
        For i = 0 To Max_SymbolCode
            SymbolStore(i) = SymbolFile.Clone(New Rectangle(i * TileSize, 0, TileSize, TileSize), Drawing.Imaging.PixelFormat.DontCare)
            SymbolStoreSelect(i) = SymbolFile.Clone(New Rectangle(i * TileSize, 16, TileSize, TileSize), Drawing.Imaging.PixelFormat.DontCare)
        Next i
        SymbolFile.Dispose()
    End Function

    Public Function MakePoint(ByVal x As Integer, ByVal y As Integer) As Point
        Dim p As Point
        p.X = x
        p.Y = y
        Return p
    End Function

    Public Function MakeNX(ByVal EntrySig As String, ByVal ExitSig As String) As String
        Dim temp As String = ""
        temp = EntrySig & "*TO*" & ExitSig
        Return temp
    End Function

    Public Interface ISignal
        Sub EnterDesignMode()
        Sub EnterOperateMode()
        Sub UpdateAspect()
        ReadOnly Property SigNumber() As String
    End Interface

    Public Interface IMisc
        Sub EnterDesignMode()
        Sub EnterTestMode()
        Sub EnterOperateMode()
        Sub Update()
        Property State() As Boolean
        ReadOnly Property ID() As String
    End Interface

End Module

' Panel Tile - One created for each tile of panel.
Public Class PanelTile
    Private SymbolCode As Integer = 156
    Private TileImage As Bitmap
    Private TileLocationRectangle As Rectangle = New Rectangle(0, 0, 16, 16)

    Public Sub New(ByVal Symbol As Integer, ByVal xLocation As Integer, ByVal yLocation As Integer)
        If xLocation >= 0 And xLocation <= Max_X And yLocation >= 0 And yLocation <= Max_Y Then
            TileLocationRectangle.X = xLocation * TileSize
            TileLocationRectangle.Y = yLocation * TileSize
        End If
        ChangeSymbol(Symbol)
    End Sub

    Public Function ChangeSymbol(ByVal Symbol As Integer) As Rectangle
        If Symbol >= 0 And Symbol <= Max_SymbolCode Then SymbolCode = Symbol
        TileImage = SymbolStore(SymbolCode).Clone
        Return TileLocationRectangle
    End Function

    Public Function TileSelected(ByVal Status As Boolean) As Rectangle
        If Status = False Then
            TileImage = SymbolStore(SymbolCode).Clone
        Else
            TileImage = SymbolStoreSelect(SymbolCode).Clone
        End If
        Return TileLocationRectangle
    End Function

    Public Sub Draw(ByVal graphicsObect As Graphics)
        graphicsObect.DrawImage(TileImage, TileLocationRectangle)
    End Sub

    Public ReadOnly Property TileLocation() As Rectangle
        Get
            Return TileLocationRectangle
        End Get
    End Property

    Public ReadOnly Property Symbol() As Integer
        Get
            Return SymbolCode
        End Get
    End Property
End Class

'Screen Elements Class Definitions

'**TOP LEVEL**
<Serializable()> Public Class ScreenElements
    Private x, y As Integer

    Public Sub New(ByVal xpos As Integer, ByVal ypos As Integer)
        If x >= 0 And x <= Max_X And y >= 0 And y <= Max_Y Then
            x = xpos
            y = ypos
        Else
            MessageBox.Show("Invalid Location", mrcc)
        End If
    End Sub

    Public Function ChangePosition(ByVal x As Integer, ByVal y As Integer)
        If x >= 0 And x <= Max_X And y >= 0 And y <= Max_Y Then
            Me.x = x
            Me.y = y
        End If
    End Function

    Public ReadOnly Property X_Pos() As Integer
        Get
            Return Me.x
        End Get
    End Property
    Public ReadOnly Property Y_Pos() As Integer
        Get
            Return Me.y
        End Get
    End Property

End Class

'**TRACK SECTION**
<Serializable()> Public Class TrackSection
    Inherits ScreenElements
    Private Reference As String = ""
    Private HasTC As Boolean = False
    Private TCAddress As Integer = -1
    Private SymbolCode As Integer = 143
    Private TrackElements As ArrayList = New ArrayList(20)

    Private TrackOcc As Boolean = False
    Private RouteSet As Boolean = False
    Private TORRFlag As Boolean = False
    Private AButtonFlag As Boolean = False

    Public Sub New(ByVal x As Integer, ByVal y As Integer, ByVal TkSecID As String, ByVal TC As Boolean, ByVal RPCInput As Integer)
        MyBase.New(x, y)
        Reference = TkSecID
        HasTC = TC
        TCAddress = RPCInput
    End Sub

    Public Sub UpdateDetails(ByVal TkSecID As String, ByVal TC As Boolean, ByVal RPCInput As Integer)
        Reference = TkSecID
        HasTC = TC
        TCAddress = RPCInput
    End Sub

    Public Function Validate() As Boolean
        Dim result As Boolean = True

        'Checks ID
        If Reference = "" Then
            result = False
            MessageBox.Show("No Track Section ID given - Please enter one", mrcc)
        Else
            'Checks for existance of Reference
            If Data.CheckReference(Reference) <> -1 Then '-1 means not found
                'Reference in use already
                Dim p As Point
                p = CType(Data.NameIndex(Reference), Point)
                If p.X <> CurrentX Or p.Y <> CurrentY Then 'checks if match is itself
                    'Name Exists and in use by another object
                    result = False
                    MessageBox.Show("Device ID " & Reference & " is already in use - Please Choose Another", mrcc)
                Else
                    'Name Exists for this object
                End If
            Else
                'Reference not in use
                'Check If device is new
                If Data.GetDeviceType(CurrentX, CurrentY) = -1 Then
                    'Device is New
                Else
                    'Device exists already
                    '(Must be TkSection or wouldn't have got here)

                    'Checks if Reference has changed
                    Dim existingobject As TrackSection = CType(Data.GetDevice(CurrentX, CurrentY), TrackSection)
                    If existingobject.Reference <> Me.Reference Then
                        'Verifies that old name has not been used in Track Elements
                        If existingobject.AttachedTrackElements Then
                            result = False
                            MessageBox.Show("Can not change ID while Track Elements still Reference the old ID", mrcc)
                        End If

                    End If
                End If
            End If
        End If
        'Validates RPC Bit
        If Not TCAddress < 0 Then
            If Data.ValidateRPCBit(TCAddress) = False Then result = False
        End If

        Return result
    End Function

    Public Sub AddTrackElement(ByVal x As Integer, ByVal y As Integer)
        If Not TrackElements.Contains(MakePoint(x, y)) Then
            TrackElements.Add(MakePoint(x, y))
        End If
    End Sub

    Public Sub RemoveTrackElement(ByVal x As Integer, ByVal y As Integer)
        If TrackElements.Contains(MakePoint(x, y)) Then TrackElements.Remove(MakePoint(x, y))
    End Sub

    Public ReadOnly Property AttachedTrackElements() As Boolean
        Get
            If TrackElements.Count = 0 Then Return False Else Return True
        End Get
    End Property

    Public ReadOnly Property TkName() As String
        Get
            Return Me.Reference
        End Get
    End Property

    Public ReadOnly Property Symbol() As Integer
        Get
            Return Me.SymbolCode
        End Get
    End Property

    Public ReadOnly Property TC() As Boolean
        Get
            Return Me.HasTC
        End Get
    End Property

    Public Property TOcc() As Boolean
        Get
            Return Me.TrackOcc
        End Get
        Set(ByVal Value As Boolean)
            If Value = True And Me.TrackOcc = False Then
                'Track Occupied
                Dim i As Integer
                For i = 0 To Me.TrackElements.Count - 1
                    Dim xy As Point = CType(TrackElements(i), Point)
                    Select Case (Data.GetDeviceType(xy.X, xy.Y))
                        Case 9 'Plain Track
                            CType(Data.GetDevice(xy.X, xy.Y), PlainTrack).OccupyTrack()
                        Case 10 'Points
                            CType(Data.GetDevice(xy.X, xy.Y), Points).OccupyTrack()
                        Case Else
                    End Select
                Next

                'Update Symbol if Test Mode
                If Mode = 1 Then
                    SymbolCode += 1
                    GUI.UpdateDisplay(Reference)
                End If

                'Enable TORR if Route Set
                If Me.USet = True And Me.AButtonFlag = False Then
                    Me.TORRAv = True
                End If
            End If

            If Value = False And Me.TrackOcc = True Then
                'Track Clear
                Dim i As Integer

                'TORR if Route Set
                If Me.USet = True And Me.TORRAv = True Then
                    TORRInProgressFlag = True
                    Me.USet = False
                    TORRInProgressFlag = False
                    Me.TORRFlag = False
                End If

                For i = 0 To Me.TrackElements.Count - 1
                    Dim xy As Point = CType(TrackElements(i), Point)
                    Select Case (Data.GetDeviceType(xy.X, xy.Y))
                        Case 9 'Plain Track
                            CType(Data.GetDevice(xy.X, xy.Y), PlainTrack).ClearTrack()
                        Case 10 'Points
                            CType(Data.GetDevice(xy.X, xy.Y), Points).ClearTrack()
                        Case Else
                    End Select
                Next

                'Update Symbol if Test Mode
                If Mode = 1 Then
                    SymbolCode -= 1
                    GUI.UpdateDisplay(Reference)
                End If

            End If
            Me.TrackOcc = Value
        End Set
    End Property

    Public Property USet() As Boolean
        Get
            Return Me.RouteSet
        End Get
        Set(ByVal Value As Boolean)
            If Value = True And Me.RouteSet = False Then
                'Route Set
                Dim i As Integer
                For i = 0 To Me.TrackElements.Count - 1
                    Dim xy As Point = CType(TrackElements(i), Point)
                    Select Case (Data.GetDeviceType(xy.X, xy.Y))
                        Case 9 'Plain Track
                            CType(Data.GetDevice(xy.X, xy.Y), PlainTrack).SubRouteSet()
                        Case 10 'Points
                            CType(Data.GetDevice(xy.X, xy.Y), Points).SubRouteSet()
                        Case Else
                    End Select
                Next
            End If

            If Value = False And Me.RouteSet = True Then
                'Clear Route
                Dim i As Integer
                For i = 0 To Me.TrackElements.Count - 1
                    Dim xy As Point = CType(TrackElements(i), Point)
                    Select Case (Data.GetDeviceType(xy.X, xy.Y))
                        Case 9 'Plain Track
                            CType(Data.GetDevice(xy.X, xy.Y), PlainTrack).SubRouteClr()
                        Case 10 'Points
                            CType(Data.GetDevice(xy.X, xy.Y), Points).SubRouteClr()
                        Case Else
                    End Select
                Next
            End If
            Me.RouteSet = Value
        End Set
    End Property

    Public Property TORRAv() As Boolean
        Get
            Return Me.TORRFlag
        End Get
        Set(ByVal Value As Boolean)
            Me.TORRFlag = Value
        End Set
    End Property

    Public Property APressed() As Boolean
        Get
            Return Me.AButtonFlag
        End Get
        Set(ByVal Value As Boolean)
            If Value = True And Me.AButtonFlag = False Then
                'Clear TORR flag
                Me.TORRFlag = False
            End If
            If Value = False And Me.AButtonFlag = True And Me.TOcc = True Then
                'Set TORR flag if train in Section
                Me.TORRFlag = True
            End If
            Me.AButtonFlag = Value
        End Set
    End Property

    Public ReadOnly Property TCInputBit() As Integer
        Get
            Return Me.TCAddress
        End Get
    End Property

    Public Sub Update()
        If Me.HasTC = True And Me.TCAddress > -1 Then
            Select Case (HW.ReadBit(Me.TCAddress))
                Case 0
                    'Occupy Track
                    Me.TOcc = True
                Case 1
                    'Clear Track
                    Me.TOcc = False
                Case Else
            End Select
        End If
    End Sub

    Public Sub EnterDesignMode()
        'Clear Track
        TOcc = False
        SymbolCode = 143 'TkSec Symbol
        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
    End Sub

    Public Sub EnterTestMode()
        'Clear Track
        If HasTC = True Then HW.SimulateInput(TCAddress, True)
        TOcc = False
        SymbolCode = 144 'TkClr Symbol
        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
    End Sub

    Public Sub EnterOperateMode()
        SymbolCode = 156 'Blank Symbol
        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
    End Sub

End Class

'**TRACK ELEMENT**
'
<Serializable()> Public Class TrackElement
    Inherits ScreenElements
    Private SymbolCode As Integer
    Private TkAssnTF As Boolean = False
    Private TkAssn As String = ""

    Public Sub New(ByVal x As Integer, ByVal y As Integer, ByVal symbol As Integer, ByVal TkAssociated As Boolean, ByVal TkSecAssn As String)
        MyBase.New(x, y)
        SymbolCode = symbol
        TkAssnTF = TkAssociated
        TkAssn = TkSecAssn

    End Sub

    Public ReadOnly Property Symbol() As Integer
        Get
            Return Me.SymbolCode
        End Get
    End Property

    Public ReadOnly Property IsTkAssn() As Boolean
        Get
            Return Me.TkAssnTF
        End Get
    End Property

    Public ReadOnly Property TkSecAssn() As String
        Get
            Return Me.TkAssn
        End Get
    End Property

    Public Sub ChangeSymbol(ByVal Symbol As Integer)
        SymbolCode = Symbol
    End Sub

    Public Sub UpdateTkAssn(ByVal TF As Boolean, ByVal TkSec As String)
        'Validated by Validate Function aready
        TkAssnTF = TF
        TkAssn = TkSec
    End Sub

    Public Function ValidateTk() As Boolean
        Dim result As Boolean = True
        If Data.CheckReference(TkAssn) <> 11 And TkAssnTF = True Then
            result = False
            MessageBox.Show("Track Section Reference is not Valid - Ensure it exists before referencing it", mrcc)
        End If
        Return result
    End Function
End Class

'**Track Element - PLAIN TRACK**
'
<Serializable()> Public Class PlainTrack
    Inherits TrackElement
    Private PointAssnTF As Boolean = False
    Private PointRef1 As String = ""
    Private PointPosRef1 As String = ""
    Private PointRef2 As String = ""
    Private PointPosRef2 As String = ""
    Private PointRef3 As String = ""
    Private PointPosRef3 As String = ""

    'Constructor
    Public Sub New(ByVal x As Integer, ByVal y As Integer, ByVal symbol As Integer, ByVal TkAssociated As Boolean, ByVal TkSecAssn As String, ByVal PointAssociated As Boolean, ByVal PRef1 As String, ByVal PPos1 As String, ByVal PRef2 As String, ByVal PPos2 As String, ByVal PRef3 As String, ByVal PPos3 As String)
        MyBase.New(x, y, symbol, TkAssociated, TkSecAssn)
        PointAssnTF = PointAssociated
        PointRef1 = PRef1
        PointPosRef1 = PPos1
        PointRef2 = PRef2
        PointPosRef2 = PPos2
        PointRef3 = PRef3
        PointPosRef3 = PPos3
    End Sub

    Public Sub UpdateSymbol(ByVal Symbol As Integer)
        If Symbol > 0 Then
            '***VALIDATE SYMBOL CODE AS APROPRIATE HERE!!
            MyBase.ChangeSymbol(Symbol)
        Else
            MessageBox.Show("Attempted to Change Symbol for Plain Track from " & Me.Symbol() & "to " & Symbol & " - Invalid", mrcc)
        End If
    End Sub

    Public Sub UpdatePointAssn(ByVal TF As Boolean, ByVal Ref1 As String, ByVal Ref1NR As String, ByVal Ref2 As String, ByVal Ref2NR As String, ByVal Ref3 As String, ByVal Ref3NR As String)
        'Validated already by Validate Function
        PointAssnTF = TF
        PointRef1 = Ref1
        PointPosRef1 = Ref1NR
        PointRef2 = Ref2
        PointPosRef2 = Ref2NR
        PointRef3 = Ref3
        PointPosRef3 = Ref3NR
    End Sub

    Public Function Validate() As Boolean
        Dim result As Boolean
        'Check Track Section Reference
        If MyBase.ValidateTk() Then result = True Else Return False

        'Check Point References
        If (Data.CheckReference(PointRef1) <> 10 Or PointPosRef1 = "") And PointAssnTF = True Then
            result = False
            MessageBox.Show("Point Association Reference 1: " & PointRef1 & " is not Valid" & vbCrLf & "Ensure it exists and has Valid Position entered before referencing it", mrcc)
        End If
        If (Data.CheckReference(PointRef2) <> 10 Or PointPosRef2 = "") And PointAssnTF = True And PointRef2 <> "" Then
            result = False
            MessageBox.Show("Point Association Reference 2: " & PointRef2 & " is not Valid" & vbCrLf & "Ensure it exists and has Valid Position entered before referencing it", mrcc)
        End If
        If (Data.CheckReference(PointRef3) <> 10 Or PointPosRef3 = "") And PointAssnTF = True And PointRef3 <> "" Then
            result = False
            MessageBox.Show("Point Association Reference 3: " & PointRef3 & " is not Valid" & vbCrLf & "Ensure it exists and has Valid Position entered before referencing it", mrcc)
        End If
        Return result
    End Function

    Public ReadOnly Property PointAssn() As Boolean
        Get
            Return PointAssnTF
        End Get
    End Property

    Public ReadOnly Property PointsRef1() As String
        Get
            Return PointRef1
        End Get
    End Property
    Public ReadOnly Property PointsRef2() As String
        Get
            Return PointRef2
        End Get
    End Property
    Public ReadOnly Property PointsRef3() As String
        Get
            Return PointRef3
        End Get
    End Property
    Public ReadOnly Property PointsPosRef1() As String
        Get
            Return PointPosRef1
        End Get
    End Property
    Public ReadOnly Property PointsPosRef2() As String
        Get
            Return PointPosRef2
        End Get
    End Property
    Public ReadOnly Property PointsPosRef3() As String
        Get
            Return PointPosRef3
        End Get
    End Property

    'Operate Mode Functions
    Public Sub SubRouteSet()
        'Displays Route Set
        If PointAssnTF = False Then
            'No Point Association
            If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                'Track Occupied
                'Leave Track Red
            Else
                'Track Clear
                'Change Track to White
                MyBase.ChangeSymbol(MyBase.Symbol + 1)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        Else
            'Point Association
            'Check 1st Point Association
            If Data.GetPoint(PointRef1).Position = PointPosRef1 Then
                If PointRef2 = "" Then
                    'No 2nd Point Association
                    If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                        'Track Occupied
                        'Leave Track Red
                    Else
                        'Track Clear
                        'Change Track to White
                        MyBase.ChangeSymbol(MyBase.Symbol + 1)
                        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                    End If
                Else
                    'Check 2nd Point Association
                    If Data.GetPoint(PointRef2).Position = PointPosRef2 Then
                        If PointRef3 = "" Then
                            'No 3rd Point Association
                            If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                                'Track Occupied
                                'Leave Track Red
                            Else
                                'Track Clear
                                'Change Track to White
                                MyBase.ChangeSymbol(MyBase.Symbol + 1)
                                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                            End If
                        Else
                            'Check 3nd Point Association
                            If Data.GetPoint(PointRef3).Position = PointPosRef3 Then
                                If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                                    'Track Occupied
                                    'Leave Track Red
                                Else
                                    'Track Clear
                                    'Change Track to White
                                    MyBase.ChangeSymbol(MyBase.Symbol + 1)
                                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                                End If
                            Else
                                'Doesn't affect this leg
                                If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                                    'Remove Existing Track Occupied Flood
                                    MyBase.ChangeSymbol(MyBase.Symbol - 2)
                                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                                End If
                            End If
                        End If
                    Else
                        'Doesn't affect this leg
                        If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                            'Remove Existing Track Occupied Flood
                            MyBase.ChangeSymbol(MyBase.Symbol - 2)
                            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                        End If
                    End If
                End If
            Else
                'Doesn't affect this leg
                If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                    'Remove Existing Track Occupied Flood
                    MyBase.ChangeSymbol(MyBase.Symbol - 2)
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            End If
        End If
    End Sub

    Public Sub SubRouteClr()
        'Displays Route Clear
        If Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
            'Track Occupied and Route Cancelled
            'Check if Track is Red
            If (MyBase.Symbol - 99) Mod 3 = 0 Then
                'Leave Red
            Else
                If (MyBase.Symbol - 98) Mod 3 = 0 Then
                    'Change White - Red
                    MyBase.ChangeSymbol(MyBase.Symbol + 1)
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                Else
                    'Change Grey - Red (Flood Track)
                    MyBase.ChangeSymbol(MyBase.Symbol + 2)
                    If TORRInProgressFlag = False Then GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            End If
        Else
            'Track Not Occupied and Route Clear
            If (MyBase.Symbol - 98) Mod 3 = 0 Then
                'Change White - Grey
                MyBase.ChangeSymbol(MyBase.Symbol - 1)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If
    End Sub

    Public Sub OccupyTrack()
        If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).USet = True Then
            If (MyBase.Symbol - 98) Mod 3 = 0 Then
                'Change White - Red
                MyBase.ChangeSymbol(MyBase.Symbol + 1)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        Else
            'Change Grey - Red (Flood Track)
            MyBase.ChangeSymbol(MyBase.Symbol + 2)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If
    End Sub

    Public Sub ClearTrack()
        'Check This Element is Red
        If (MyBase.Symbol - 99) Mod 3 = 0 Then
            If Data.GetTKSection(MyBase.TkSecAssn).USet = True Then
                If PointAssnTF = True Then
                    If Data.GetPoint(PointRef1).Position = PointPosRef1 Then
                        If PointRef2 = "" Then
                            'No 2nd Point Reference
                            'Change Red - White
                            MyBase.ChangeSymbol(MyBase.Symbol - 1)
                            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                        Else
                            If Data.GetPoint(PointRef2).Position = PointPosRef2 Then
                                'Change Red - White
                                MyBase.ChangeSymbol(MyBase.Symbol - 1)
                                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                            Else
                                'Not on Set Route
                                'Change Red - Grey
                                MyBase.ChangeSymbol(MyBase.Symbol - 2)
                                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                            End If
                        End If
                    Else
                        'Not on Set Route
                        'Change Red - Grey
                        MyBase.ChangeSymbol(MyBase.Symbol - 2)
                        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                    End If
                Else
                    'No Point Association
                    'Change Red - White
                    MyBase.ChangeSymbol(MyBase.Symbol - 1)
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            Else
                'Change Red - Grey
                MyBase.ChangeSymbol(MyBase.Symbol - 2)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If
    End Sub
End Class

'**Track Element - POINTS**
'
<Serializable()> Public Class Points
    Inherits TrackElement
    Private Reference As String = ""
    Private RPCOutputBit As Integer = -1
    Private CurrentPos As String = "N"
    Private KeySetting As String = "C"
    Private PointAssnTF As Boolean = False
    Private PointRef1 As String = ""
    Private PointPosRef1 As String = ""
    Private PointRef2 As String = ""
    Private PointPosRef2 As String = ""
    Private WorksWithRef As String = ""
    Private SwapNR As Boolean = False
    Private PointAssns As ArrayList = New ArrayList(20)
    Private ILMiscFalse As String = ""
    Private ILMiscTrue As String = ""

    'Constructor
    Public Sub New(ByVal x As Integer, ByVal y As Integer, ByVal symbol As Integer, ByVal TkAssociated As Boolean, ByVal TkSecAssn As String, ByVal PointName As String, ByVal RPCOut As Integer, ByVal PointAssociated As Boolean, ByVal PRef1 As String, ByVal PPos1 As String, ByVal PRef2 As String, ByVal PPos2 As String, ByVal WorksWith As String, ByVal SwapNmlRev As Boolean, ByVal MiscTrue As String, ByVal MiscFalse As String)
        MyBase.New(x, y, symbol, TkAssociated, TkSecAssn)
        Reference = PointName
        RPCOutputBit = RPCOut
        If Not RPCOutputBit < 0 Then Data.RegisterRPCBit(RPCOut)
        CurrentPos = "N" 'Defaults to Normal
        KeySetting = "C" 'Defaults to Centre
        PointAssnTF = PointAssociated
        PointRef1 = PRef1
        PointPosRef1 = PPos1
        PointRef2 = PRef2
        PointPosRef2 = PPos2
        WorksWithRef = WorksWith
        SwapNR = SwapNmlRev
        ILMiscTrue = MiscTrue
        ILMiscFalse = MiscFalse
    End Sub

    Public Sub UpdateSymbol(ByVal Symbol As Integer)
        If Symbol > 0 Then
            '***VALIDATE SYMBOL CODE AS APROPRIATE HERE!!
            MyBase.ChangeSymbol(Symbol)
        Else
            MessageBox.Show("Attempted to Change Symbol for Point Track from " & Me.Symbol() & "to " & Symbol & " - Invalid", mrcc)
        End If
    End Sub

    Public Sub UpdatePointAssn(ByVal TF As Boolean, ByVal Ref1 As String, ByVal Ref1NR As String, ByVal Ref2 As String, ByVal Ref2NR As String, ByVal Pair As String)
        'Validated already by Validate Function
        PointAssnTF = TF
        PointRef1 = Ref1
        PointPosRef1 = Ref1NR
        PointRef2 = Ref2
        PointPosRef2 = Ref2NR
        WorksWithRef = Pair
    End Sub

    Public Sub UpdatePoint(ByVal PointName As String, ByVal RPCBit As Integer, ByVal NRSwap As Boolean, ByVal MiscTrue As String, ByVal MiscFalse As String)
        'Validated already by Validate Function
        Reference = PointName
        RPCOutputBit = RPCBit
        SwapNR = NRSwap
        ILMiscTrue = MiscTrue
        ILMiscFalse = MiscFalse
    End Sub

    Public ReadOnly Property MiscTrue() As String
        Get
            Return ILMiscTrue
        End Get
    End Property

    Public ReadOnly Property MiscFalse() As String
        Get
            Return ILMiscFalse
        End Get
    End Property
    Public ReadOnly Property PointName() As String
        Get
            Return Reference
        End Get
    End Property

    Public ReadOnly Property OutputBit() As Integer
        Get
            Return RPCOutputBit
        End Get
    End Property

    Public ReadOnly Property NRInvert() As Boolean
        Get
            Return SwapNR
        End Get
    End Property

    Public Property PointPair() As String
        Get
            Return WorksWithRef
        End Get

        Set(ByVal Value As String)
            WorksWithRef = Value
        End Set
    End Property

    Public ReadOnly Property Position() As String
        Get
            Return CurrentPos
        End Get
    End Property

    Public ReadOnly Property Key() As String
        Get
            Return KeySetting
        End Get
    End Property

    Public ReadOnly Property PointAssn() As Boolean
        Get
            Return PointAssnTF
        End Get
    End Property

    Public ReadOnly Property PointAssnRef1() As String
        Get
            Return PointRef1
        End Get
    End Property

    Public ReadOnly Property PointAssnPos1() As String
        Get
            Return PointPosRef1
        End Get
    End Property

    Public ReadOnly Property PointAssnRef2() As String
        Get
            Return PointRef2
        End Get
    End Property

    Public ReadOnly Property PointAssnPos2() As String
        Get
            Return PointPosRef2
        End Get
    End Property

    Public Sub AddTrackElement(ByVal x As Integer, ByVal y As Integer)
        If Not PointAssns.Contains(MakePoint(x, y)) Then
            PointAssns.Add(MakePoint(x, y))
        End If
    End Sub

    Public Sub RemoveTrackElement(ByVal x As Integer, ByVal y As Integer)
        If PointAssns.Contains(MakePoint(x, y)) Then PointAssns.Remove(MakePoint(x, y))
    End Sub

    Public ReadOnly Property AttachedTrackElements() As Boolean
        Get
            If PointAssns.Count = 0 Then Return False Else Return True
        End Get
    End Property

    Public Function Validate() As Boolean
        Dim result As Boolean = True
        'Check Point Name
        'Checks for existance of Reference
        If Reference = "" Then
            result = False
            MessageBox.Show("No Point ID given - Please Enter One", mrcc)
        Else
            If Data.CheckReference(Reference) <> -1 Then '-1 means not found
                'Reference in use already
                Dim p As Point
                p = CType(Data.NameIndex(Reference), Point)
                If p.X <> CurrentX Or p.Y <> CurrentY Then 'checks if match is itself
                    'Name Exists and in use by another object
                    result = False
                    MessageBox.Show("Device ID " & Reference & " is already in use - Please Choose Another", mrcc)
                Else
                    'Name Exists for this object
                End If
            Else
                'Reference not in use
                'Check If device is new
                If Data.GetDeviceType(CurrentX, CurrentY) = -1 Then
                    'Device is New
                Else
                    'Device exists already
                    '(Must be Points or wouldn't have got here)
                    Dim existingobject As Points = CType(Data.GetDevice(CurrentX, CurrentY), Points)
                    'Checks if Reference has changed
                    If existingobject.Reference <> Me.Reference Then
                        'Verifies that old name has not been used in Track Elements
                        If existingobject.AttachedTrackElements Then
                            result = False
                            MessageBox.Show("Can not change ID while Track Elements still Reference the old ID", mrcc)
                        End If

                    End If
                End If
            End If
        End If
        'Check Track References
        If MyBase.ValidateTk() = False Then result = False

        'Check RPC Reference
        If Not RPCOutputBit < 0 Then
            If Data.ValidateRPCBit(RPCOutputBit) = False Then result = False
        End If

        'Check Point References
        If PointAssnTF = True Then
            If (Data.CheckReference(PointRef1) <> 10 Or PointPosRef1 = "") Then
                result = False
                MessageBox.Show("PointAssociation Reference 1: " & PointRef1 & " is not Valid - Ensure it exists and has Valid position before referencing it", mrcc)
            Else
                'Reference OK - Check that it is not referencing itself!
                If PointRef1 = PointName Then
                    result = False
                    MessageBox.Show("Point Association Reference 1 is not Valid - Can not Reference itself!", mrcc)
                End If
            End If

            If PointRef2 <> "" Then
                If (Data.CheckReference(PointRef2) <> 10 Or PointPosRef2 = "") Then
                    result = False
                    MessageBox.Show("PointAssociation Reference 2: " & PointRef2 & " is not Valid - Ensure it exists and has Valid position before referencing it", mrcc)
                Else
                    'Reference OK - Check that it is not referencing itself!
                    If PointRef2 = PointName Then
                        result = False
                        MessageBox.Show("Point Association Reference 2 is not Valid - Can not Reference itself!", mrcc)
                    End If
                End If
            End If
        End If

        'Check WorksWith Reference
        If WorksWithRef <> "" Then
            If Data.CheckReference(WorksWithRef) <> 10 Then
                result = False
                MessageBox.Show("Works With Point Reference: " & WorksWithRef & " is not Valid - Ensure it exists before referencing it", mrcc)
            Else
                'Reference OK - Check that it is not referencing itself!
                If WorksWithRef = PointName Then
                    result = False
                    MessageBox.Show("Point Works With Reference is not Valid - Can not Reference itself!", mrcc)
                End If

                'Check Other Point hasn't been Paired already
                If Not CType(Data.GetObject(WorksWithRef), Points).WorksWithRef = "" Then
                    Dim AlreadyPaired As Boolean = True

                    'Check that it isn't a name change of this Point Object
                    If Data.GetDeviceType(CurrentX, CurrentY) = -1 Then
                        'skip this check - New Object
                    Else
                        Dim existing As Points = CType(Data.GetDevice(CurrentX, CurrentY), Points)
                        If CType(Data.GetObject(WorksWithRef), Points).WorksWithRef = existing.PointName Then
                            'Point Free to Pair but Reference needs updating
                            AlreadyPaired = False
                        End If
                    End If
                    If AlreadyPaired = True Then
                        'Already Paired - Error
                        result = False
                        MessageBox.Show("Point " & WorksWithRef & " is already Paired to another Point - Can Only Pair with One Point", mrcc)
                    End If
                End If
            End If
        End If
        Return result
    End Function

    'Validate Interlocking Data
    Public Function ILDataValid() As Boolean
        If ValidateILData() = "" Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Function ValidateILData() As String
        Dim errorreport As String = ""
        If ILMiscTrue <> "" Then
            If Not ((Data.CheckReference(ILMiscTrue) = 12) Or (Data.CheckReference(ILMiscTrue) = 14)) Then errorreport &= "Point " & Reference & ": References Non Existant Misc. Device " & ILMiscTrue & vbCrLf
        End If
        If ILMiscFalse <> "" Then
            If Not ((Data.CheckReference(ILMiscFalse) = 12) Or (Data.CheckReference(ILMiscFalse) = 14)) Then errorreport &= "Point " & Reference & ": References Non Existant Misc. Device " & ILMiscFalse & vbCrLf
        End If
        Return errorreport
    End Function

    'Operate Mode Functions
    Public Sub SubRouteSet()
        'Displays Route Set
        If PointAssnTF = False Then
            'No Point Association
            If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                'Track Occupied
                'Leave Track Red
            Else
                'Track Clear
                'Change Track to White
                MyBase.ChangeSymbol(MyBase.Symbol + 2)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        Else
            'Point Association
            'Check Point Association
            If Data.GetPoint(PointRef1).Position = PointPosRef1 Then
                If PointRef2 = "" Then
                    'No 2nd Point Association
                    If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                        'Track Occupied
                        'Leave Track Red
                    Else
                        'Track Clear
                        'Change Track to White
                        MyBase.ChangeSymbol(MyBase.Symbol + 2)
                        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                    End If
                Else
                    'Check 2nd Point Association
                    If Data.GetPoint(PointRef2).Position = PointPosRef2 Then
                        If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                            'Track Occupied
                            'Leave Track Red
                        Else
                            'Track Clear
                            'Change Track to White
                            MyBase.ChangeSymbol(MyBase.Symbol + 2)
                            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                        End If
                    Else
                        'Doesn't affect this leg
                        If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                            'Remove Existing Track Occupied Flood
                            MyBase.ChangeSymbol(MyBase.Symbol - 3)
                            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                        End If
                    End If
                End If
            Else
                'Doesn't affect this leg
                If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
                    'Remove Existing Track Occupied Flood
                    MyBase.ChangeSymbol(MyBase.Symbol - 3)
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            End If
        End If
    End Sub

    Public Sub SubRouteClr()
        'Displays Route Clear
        If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).TOcc = True Then
            'Track Occupied and Route Cancelled
            'Check if Track is Red
            If (MyBase.Symbol - 35) Mod 4 = 0 Then
                'Flood Red
                MyBase.ChangeSymbol(MyBase.Symbol - 2)
                If TORRInProgressFlag = False Then GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            Else
                If (MyBase.Symbol - 94) Mod 4 = 0 Then
                    'Change White - Red (Flood Red)
                    MyBase.ChangeSymbol(MyBase.Symbol - 1)
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                Else
                    'Change Grey - Red (Flood Track)
                    MyBase.ChangeSymbol(MyBase.Symbol + 1)
                    If TORRInProgressFlag = False Then GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            End If
        Else
            'Track Not Occupied and Route Clear
            If (MyBase.Symbol - 34) Mod 4 = 0 Then
                'Change White - Grey
                MyBase.ChangeSymbol(MyBase.Symbol - 2)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If
    End Sub

    Public Sub OccupyTrack()
        If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).USet = True Then
            'Route Set
            If (MyBase.Symbol - 34) Mod 4 = 0 Then
                'Change White - Red
                MyBase.ChangeSymbol(MyBase.Symbol + 1)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        Else
            'Change Grey - Red (Flood Track)
            MyBase.ChangeSymbol(MyBase.Symbol + 1)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If
    End Sub

    Public Sub ClearTrack()
        'Check This Element is Red
        If (MyBase.Symbol - 35) Mod 2 = 0 Then
            If MyBase.IsTkAssn = True And Data.GetTKSection(MyBase.TkSecAssn).USet = True Then
                'Route Set
                If PointAssnTF = True Then
                    If Data.GetPoint(PointRef1).Position = PointPosRef1 Then
                        'Change Red - White
                        MyBase.ChangeSymbol(MyBase.Symbol - 1)
                        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                    Else
                        'Not on Set Route
                        'Change Red - Grey
                        MyBase.ChangeSymbol(MyBase.Symbol - 3)
                        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                    End If
                Else
                    'No Point Association
                    'Change Red - White
                    MyBase.ChangeSymbol(MyBase.Symbol - 1)
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            Else
                'Change Red (Flood) - Grey
                MyBase.ChangeSymbol(MyBase.Symbol - 1)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If
    End Sub

    Public Sub MoveNormal()
        If CurrentPos = "R" Then
            'Update Symbol
            If SwapNR = True Then
                MyBase.ChangeSymbol(MyBase.Symbol + 4)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            Else
                MyBase.ChangeSymbol(MyBase.Symbol - 4)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If

            'Change Bit in Hardware
            If RPCOutputBit <> -1 Then
                HW.ClearBit(RPCOutputBit)
            End If
        End If
        CurrentPos = "N"
    End Sub

    Public Sub MoveReverse()
        If CurrentPos = "N" Then
            'Update Symbol
            If SwapNR = True Then
                MyBase.ChangeSymbol(MyBase.Symbol - 4)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            Else
                MyBase.ChangeSymbol(MyBase.Symbol + 4)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If

            'Change Bit in Hardware
            If RPCOutputBit <> -1 Then
                HW.SetBit(RPCOutputBit)
            End If
        End If
        CurrentPos = "R"
    End Sub

    Public Function KeyNormal() As String
        Dim FeedbackText As String = ""
        Dim errorreport As String = CallNormal("Key")
        If errorreport = "" Then
            KeySetting = "N"
        Else
            FeedbackText = "Unable to Key Point " & Reference & " Normal: -" & vbCrLf & vbCrLf & errorreport
        End If
        Return FeedbackText
    End Function

    Public Function KeyReverse() As String
        Dim FeedbackText As String = ""
        Dim errorreport As String = CallReverse("Key")
        If errorreport = "" Then
            KeySetting = "R"
        Else
            FeedbackText = "Unable to Key Point " & Reference & " Reverse: -" & vbCrLf & vbCrLf & errorreport
        End If
        Return FeedbackText
    End Function

    Public Sub KeyCentre()
        KeySetting = "C"
    End Sub

    Public Function CallNormal(ByVal Sender As String) As String
        Dim errorreport As String = ""
        If CurrentPos = "N" Then
            'Already Set
        Else
            'Check No Route Set
            If MyBase.IsTkAssn = True Then
                If Data.GetTKSection(MyBase.TkSecAssn).USet = True Then
                    'Reject - Routes Set
                    errorreport &= "Unable to Call Point " & Reference & " Normal : Conflicting Route is set." & vbCrLf
                End If
            End If
            'Check Misc Conditions are met
            If ILMiscTrue <> "" Then
                If Data.GetMiscState(ILMiscTrue) = False Then
                    errorreport &= "Unable to Call Point " & Reference & " Normal : Misc. Condition " & ILMiscTrue & " is False." & vbCrLf
                End If
            End If
            If ILMiscFalse <> "" Then
                If Data.GetMiscState(ILMiscFalse) = True Then
                    errorreport &= "Unable to Call Point " & Reference & " Normal : Misc. Condition " & ILMiscFalse & " is True." & vbCrLf
                End If
            End If

            'Check PointKey
            If KeySetting <> "C" And Sender <> "Key" Then
                errorreport &= "Unable to Call Point " & Reference & " Normal : Points Key is Not Centred." & vbCrLf
            End If

            'Call Point Pair and ensure it moves
            If errorreport = "" And WorksWithRef <> "" And Sender <> "Pair" Then
                errorreport &= Data.GetPoint(WorksWithRef).CallNormal("Pair")
            End If

            If errorreport = "" Then
                'Move Points if no Errors
                MoveNormal()
            End If
        End If

        Return errorreport
    End Function

    Public Function CallReverse(ByVal Sender As String) As String
        Dim errorreport As String = ""
        If CurrentPos = "R" Then
            'Already Set
        Else
            'Check No Route Set
            If MyBase.IsTkAssn = True Then
                If Data.GetTKSection(MyBase.TkSecAssn).USet = True Then
                    'Reject - Routes Set
                    errorreport &= "Unable to Call Point " & Reference & " Reverse : Conflicting Route is set." & vbCrLf
                End If
            End If
            'Check Misc Conditions are met
            If ILMiscTrue <> "" Then
                If Data.GetMiscState(ILMiscTrue) = False Then
                    errorreport &= "Unable to Call Point " & Reference & " Reverse : Misc. Condition " & ILMiscTrue & " is False." & vbCrLf
                End If
            End If
            If ILMiscFalse <> "" Then
                If Data.GetMiscState(ILMiscFalse) = True Then
                    errorreport &= "Unable to Call Point " & Reference & " Reverse : Misc. Condition " & ILMiscFalse & " is True." & vbCrLf
                End If
            End If

            'Check PointKey
            If KeySetting <> "C" And Sender <> "Key" Then
                errorreport &= "Unable to Call Point " & Reference & " Reverse : Points Key is Not Centred." & vbCrLf
            End If

            'Call Point Pair and ensure it moves
            If errorreport = "" And WorksWithRef <> "" And Sender <> "Pair" Then
                errorreport &= Data.GetPoint(WorksWithRef).CallReverse("Pair")
            End If

            If errorreport = "" Then
                'Move Points if no Errors
                MoveReverse()
            End If
        End If

        Return errorreport
    End Function

End Class

'SIGNAL
'******

<Serializable()> Public Class Signal
    Inherits ScreenElements

    Private SignalNumber As String = ""
    Private DAspect As Integer = 0
    Private Aspect As Integer = 0
    Private SymbolCode As Integer = 0
    Public RoutesFromSig As New ArrayList(20)
    'Additional Aspect Conditions
    Public MiscTrue As String = ""
    Public MiscFalse As String = ""
    Public AutoSig As Boolean = False

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal ID As String, ByVal DefaultAspect As Integer, ByVal SigSymbol As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAuto As Boolean)
        MyBase.New(x, y)

        SignalNumber = ID
        DAspect = DefaultAspect
        SymbolCode = SigSymbol
        RoutesFromSig = Routes
        MiscTrue = ILMiscTrue
        MiscFalse = ILMiscFalse
        AutoSig = ILAuto
    End Sub

    Friend Sub UpdateSigDetails(ByVal SigID As String, ByVal DefAspect As Integer, ByVal Symbol As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAuto As Boolean)
        'Validated by Form
        SignalNumber = SigID
        DAspect = DefAspect
        SymbolCode = Symbol
        RoutesFromSig = Routes
        MiscTrue = ILMiscTrue
        MiscFalse = ILMiscFalse
        AutoSig = ILAuto
    End Sub

    Public Sub ChangeSymbol(ByVal symbol As Integer)
        SymbolCode = symbol
    End Sub

    Public Function ValidateSig() As Boolean
        Dim result As Boolean = True

        ' Check Signal ID
        If SignalID = "" Then
            result = False
            MessageBox.Show("No Signal ID given - Please Enter One", mrcc)
        Else
            If Data.CheckReference(SignalID) <> -1 Then '-1 means not found
                'Reference in use already
                Dim p As Point
                p = CType(Data.NameIndex(SignalID), Point)
                If p.X <> CurrentX Or p.Y <> CurrentY Then 'checks if match is itself
                    'Name Exists and in use by another object
                    result = False
                    MessageBox.Show("Device ID " & SignalID & " is already in use - Please Choose Another", mrcc)
                Else
                    'Name Exists for this object
                End If
            Else
                'Reference not in use
            End If
        End If

        'Check Default Aspect is set and valid
        If DAspect < 0 Or DAspect > 3 Then
            'Default Aspect Invalid
            result = False
            MessageBox.Show("Default Aspect has not been set", mrcc)
        End If

        Return result

    End Function

    Public Function ILDataValid() As Boolean
        If ValidateILData() = "" Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Function ValidateILData() As String
        Dim i As Integer
        Dim errorreport As String = ""
        For i = 1 To RoutesFromSig.Count
            If Not Data.CheckRouteExists(RoutesFromSig(i - 1)) Then
                errorreport &= "Signal " & SignalNumber & ": References Non Existant Route " & RoutesFromSig(i - 1) & vbCrLf
            End If
        Next
        If RoutesFromSig.Count > 1 And AutoSig = True Then errorreport &= "Signal " & SignalNumber & ":Auto Signal has more than 1 Route in its Aspect Conditions" & vbCrLf
        If RoutesFromSig.Count < 1 And AutoSig = True Then errorreport &= "Signal " & SignalNumber & ":Auto Signal has no Route in its Aspect Conditions" & vbCrLf
        If errorreport = "" And AutoSig = True Then
            'Create Auto Route References
            Data.GetRoute(RoutesFromSig(0)).AButtonRef = "**AUTO**"
        End If
        If MiscTrue <> "" Then
            If Not (Data.CheckReference(MiscTrue) = 12 Or Data.CheckReference(MiscTrue) = 14) Then
                errorreport &= "Signal " & SignalNumber & ": Misc. True Reference, " & MiscTrue & " is not a Valid Misc. Device." & vbCrLf
            End If
        End If
        If MiscFalse <> "" Then
            If Not (Data.CheckReference(MiscFalse) = 12 Or Data.CheckReference(MiscFalse) = 14) Then
                errorreport &= "Signal " & SignalNumber & ": Misc. False Reference, " & MiscFalse & " is not a Valid Misc. Device." & vbCrLf
            End If
        End If
        Return errorreport
    End Function

    Public ReadOnly Property SignalID() As String
        Get
            Return SignalNumber
        End Get
    End Property

    Public ReadOnly Property SigDefAspect() As Integer
        Get
            Return DAspect
        End Get
    End Property

    Public Property SigAspect() As Integer

        Get
            Return Aspect
        End Get
        Set(ByVal Value As Integer)
            If Value >= 0 And Value <= 3 Then Aspect = Value
        End Set
    End Property

    Public ReadOnly Property Symbol() As Integer
        Get
            Return SymbolCode
        End Get
    End Property

    'OPERATE MODE FUNCTIONs

    Public Function GetNextSigAspect() As Integer
        'Checks to see if a route is set and Misc Conditions Met
        'Then checks if Tracks are clear within set route
        'Return -1 for No Route Set/Misc Conditions Failed/ Tracks Not Clear
        'Else Return Aspect Code of Exit Signal of Set Route
        'Multiple Routes may be set if Conflicting Routes not entered
        'Takes last Route in Route List found to be Set
        Dim i As Integer
        Dim result As Integer = -1
        'Check If Route is set

        For i = 0 To RoutesFromSig.Count - 1
            If Data.GetRoute(RoutesFromSig(i)).USet = True Then
                If Data.GetRoute(RoutesFromSig(i)).TracksClear = True Then
                    Dim nextsig As String = Data.GetRoute(RoutesFromSig(i)).ExitSig
                    Dim type As Integer = Data.CheckReference(nextsig)
                    Select Case (type)
                        Case 0 'Shunt
                            result = CType(Data.GetByName(nextsig), SignalShunt).Aspect
                        Case 2 '2 Aspect
                            result = CType(Data.GetByName(nextsig), Signal2Aspect).Aspect
                        Case 4 '3 Aspect
                            result = CType(Data.GetByName(nextsig), Signal3Aspect).Aspect
                        Case 6 '4 Aspect
                            result = CType(Data.GetByName(nextsig), Signal4Aspect).Aspect
                        Case 8 'Isolated Exit
                            result = CType(Data.GetByName(nextsig), IsolatedExit).Aspect
                    End Select
                End If
            End If
        Next

        'Check Misc Conditions
        If MiscTrue <> "" Then
            If Data.GetMiscState(MiscTrue) = False Then result = -1
        End If

        If MiscFalse <> "" Then
            If Data.GetMiscState(MiscFalse) = True Then result = -1
        End If

        Return result
    End Function
End Class

'Signal - SHUNT SIGNAL
<Serializable()> Public Class SignalShunt
    Inherits Signal
    Implements ISignal

    Private Multiplex As Boolean = False
    Private RPCBit1 As Integer = -1
    Private RPCBit2 As Integer = -1
    Private NoRed As Boolean = False

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal ID As String, ByVal DefaultAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPC1 As Integer, ByVal RPC2 As Integer, ByVal MainAssnTF As Boolean, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.New(x, y, ID, DefaultAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPC1
        RPCBit2 = RPC2
        NoRed = MainAssnTF
    End Sub

    Public ReadOnly Property SignalNumber() As String Implements ISignal.SigNumber
        Get
            Return MyBase.SignalID
        End Get
    End Property

    Public ReadOnly Property Multiplexed() As Boolean
        Get
            Return Multiplex
        End Get
    End Property

    Public ReadOnly Property MainSigAssn() As Boolean
        Get
            Return NoRed
        End Get
    End Property

    Public ReadOnly Property RPC1() As Integer
        Get
            Return RPCBit1
        End Get
    End Property

    Public ReadOnly Property RPC2() As Integer
        Get
            Return RPCBit2
        End Get
    End Property

    'Validate
    Public Function validate() As Boolean
        Dim result As Boolean = True
        If MyBase.ValidateSig = False Then result = False

        'Check RPC Reference 1
        If Not RPCBit1 < 0 Then
            If Data.ValidateRPCBit(RPCBit1) = False Then result = False
        End If

        'Check RPC Reference 2
        If Not RPCBit2 < 0 Then
            If Data.ValidateRPCBit(RPCBit2) = False Then result = False
            If Multiplex = True Then
                'Non-Essential RPC Bit Allocated
                result = False
                MessageBox.Show("RPC Address 2 is not required with Multiplexed Operation - Please Set to -1", mrcc)
            End If
        End If
        Return result
    End Function

    'Update Details
    Public Sub UpdateDetails(ByVal SigID As String, ByVal DefAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPCRef1 As Integer, ByVal RPCRef2 As Integer, ByVal MainAssnTF As Boolean, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.UpdateSigDetails(SigID, DefAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPCRef1
        RPCBit2 = RPCRef2
        NoRed = MainAssnTF

    End Sub

    'OPERATE MODE FUNCTIONS

    Public Overridable Sub UpdateAspect() Implements ISignal.UpdateAspect
        Dim NextAspect As Integer = MyBase.GetNextSigAspect
        Select Case (NextAspect)
            Case -1 'No Route Set or Misc Conditions Failed
                'Change to Red
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Already Stop
                    Case 1
                        'Proceed - Stop
                        If NoRed = False Then
                            'White - Red
                            MyBase.ChangeSymbol(MyBase.Symbol - 1)
                        Else
                            'White - Grey
                            MyBase.ChangeSymbol(MyBase.Symbol - 2)
                        End If
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case Else
                        MessageBox.Show("Signal, " & MyBase.SignalID & " Invalid Aspect Code - " & MyBase.SigAspect, mrcc)
                End Select
                MyBase.SigAspect = 0
                If Multiplex = False Then
                    'Assume 1=Stop,2=Proceed
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                End If

            Case Else
                'Change to Proceed
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Stop - Proceed
                        If NoRed = False Then
                            'Red - White
                            MyBase.ChangeSymbol(MyBase.Symbol + 1)
                        Else
                            'Grey - White
                            MyBase.ChangeSymbol(MyBase.Symbol + 2)
                        End If
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 1
                        'Already Proceed
                    Case Else
                        MessageBox.Show("Signal, " & MyBase.SignalID & " Invalid Aspect Code - " & MyBase.SigAspect, mrcc)
                End Select
                MyBase.SigAspect = 1
                If Multiplex = False Then
                    'Assume 1=Stop,2=Proceed
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                End If
        End Select
    End Sub

    Public Overridable Sub EnterDesignMode() Implements ISignal.EnterDesignMode
        'Symbol to Grey
        Select Case (MyBase.SigAspect)
            Case 0
                If NoRed = False Then
                    'Red - Grey
                    MyBase.ChangeSymbol(MyBase.Symbol - 1)
                End If
            Case 1
                'White - Gray
                MyBase.ChangeSymbol(MyBase.Symbol - 2)
        End Select
        GUI.UpdateDisplay(MyBase.SignalID)
        'Aspect to Stop
        MyBase.SigAspect = 0
    End Sub

    Public Overridable Sub EnterOperateMode() Implements ISignal.EnterOperateMode
        MyBase.SigAspect = MyBase.SigDefAspect
        If MyBase.SigAspect = 0 And NoRed = False Then MyBase.ChangeSymbol(MyBase.Symbol + 1)
        If MyBase.SigAspect = 1 Then MyBase.ChangeSymbol(MyBase.Symbol + 2)
        GUI.UpdateDisplay(MyBase.SignalID)
        'HW set by first cycle of Update
    End Sub
End Class

'Signal - TWO ASPECT SIGNAL
<Serializable()> Public Class Signal2Aspect
    Inherits Signal
    Implements ISignal

    Private Multiplex As Boolean = False
    Private RPCBit1 As Integer = -1
    Private RPCBit2 As Integer = -1

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal ID As String, ByVal DefaultAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPC1 As Integer, ByVal RPC2 As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.New(x, y, ID, DefaultAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPC1
        RPCBit2 = RPC2
    End Sub


    Public ReadOnly Property SignalNumber() As String Implements ISignal.SigNumber
        Get
            Return MyBase.SignalID
        End Get
    End Property

    Public ReadOnly Property Multiplexed() As Boolean
        Get
            Return Multiplex
        End Get
    End Property

    Public ReadOnly Property RPC1() As Integer
        Get
            Return RPCBit1
        End Get
    End Property

    Public ReadOnly Property RPC2() As Integer
        Get
            Return RPCBit2
        End Get
    End Property

    'Validate
    Public Function validate() As Boolean
        Dim result As Boolean = True
        If MyBase.ValidateSig = False Then result = False

        'Check RPC Reference 1
        If Not RPCBit1 < 0 Then
            If Data.ValidateRPCBit(RPCBit1) = False Then result = False
        End If

        'Check RPC Reference 2
        If Not RPCBit2 < 0 Then
            If Data.ValidateRPCBit(RPCBit2) = False Then result = False
            If Multiplex = True Then
                'Non-Essential RPC Bit Allocated
                result = False
                MessageBox.Show("RPC Address 2 is not required with Multiplexed Operation - Please Set to -1", mrcc)
            End If
        End If
        Return result
    End Function

    'Update Details
    Public Sub UpdateDetails(ByVal SigID As String, ByVal DefAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPCRef1 As Integer, ByVal RPCRef2 As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.UpdateSigDetails(SigID, DefAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPCRef1
        RPCBit2 = RPCRef2
    End Sub

    'OPERATE MODE FUNCTIONS

    Public Overridable Sub UpdateAspect() Implements ISignal.UpdateAspect
        Dim NextAspect As Integer = MyBase.GetNextSigAspect
        Select Case (NextAspect)
            Case -1 'No Route Set or Misc Conditions Failed
                'Change to Red
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Already Red
                    Case 3
                        'Green - Red
                        MyBase.ChangeSymbol(MyBase.Symbol - 3)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case Else
                        MessageBox.Show("Signal, " & MyBase.SignalID & " Invalid Aspect Code - " & MyBase.SigAspect, mrcc)
                End Select
                MyBase.SigAspect = 0
                If Multiplex = False Then
                    'Assume 1=Red,2=Green
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                End If

            Case Else
                'Change to Green
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Red - Green
                        MyBase.ChangeSymbol(MyBase.Symbol + 3)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 3
                        'Already Green
                    Case Else
                        MessageBox.Show("Signal, " & MyBase.SignalID & " Invalid Aspect Code - " & MyBase.SigAspect, mrcc)
                End Select
                MyBase.SigAspect = 3
                If Multiplex = False Then
                    'Assume 1=Red,2=Green
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                End If
        End Select
    End Sub

    Public Overridable Sub EnterDesignMode() Implements ISignal.EnterDesignMode
        'Symbol to Grey
        Select Case (MyBase.SigAspect)
            Case 0
                'Red - Grey
                MyBase.ChangeSymbol(MyBase.Symbol - 2)
                GUI.UpdateDisplay(MyBase.SignalID)
            Case 3
                'Green - Gray
                MyBase.ChangeSymbol(MyBase.Symbol - 5)
                GUI.UpdateDisplay(MyBase.SignalID)
        End Select
        'Aspect to Red
        MyBase.SigAspect = 0
    End Sub

    Public Overridable Sub EnterOperateMode() Implements ISignal.EnterOperateMode
        MyBase.SigAspect = MyBase.SigDefAspect
        MyBase.ChangeSymbol(MyBase.Symbol + MyBase.SigAspect + 2)
        GUI.UpdateDisplay(MyBase.SignalID)
        'HW set by first cycle of Update
    End Sub
End Class

'SIGNAL 3 ASPECT
'
<Serializable()> Public Class Signal3Aspect
    Inherits Signal
    Implements ISignal

    Private Multiplex As Boolean = False
    Private RPCBit1 As Integer = -1
    Private RPCBit2 As Integer = -1
    Private RPCBit3 As Integer = -1

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal ID As String, ByVal DefaultAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPC1 As Integer, ByVal RPC2 As Integer, ByVal RPC3 As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.New(x, y, ID, DefaultAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPC1
        RPCBit2 = RPC2
        RPCBit3 = RPC3
    End Sub


    Public ReadOnly Property SignalNumber() As String Implements ISignal.SigNumber
        Get
            Return MyBase.SignalID
        End Get
    End Property

    Public ReadOnly Property Multiplexed() As Boolean
        Get
            Return Multiplex
        End Get
    End Property

    Public ReadOnly Property RPC1() As Integer
        Get
            Return RPCBit1
        End Get
    End Property

    Public ReadOnly Property RPC2() As Integer
        Get
            Return RPCBit2
        End Get
    End Property

    Public ReadOnly Property RPC3() As Integer
        Get
            Return RPCBit3
        End Get
    End Property

    'Validate
    Public Function validate() As Boolean
        Dim result As Boolean = True
        If MyBase.ValidateSig = False Then result = False

        'Check RPC Reference 1
        If Not RPCBit1 < 0 Then
            If Data.ValidateRPCBit(RPCBit1) = False Then result = False
        End If

        'Check RPC Reference 2
        If Not RPCBit2 < 0 Then
            If Data.ValidateRPCBit(RPCBit2) = False Then result = False
        End If

        'Check RPC Reference 3
        If Not RPCBit3 < 0 Then
            If Data.ValidateRPCBit(RPCBit3) = False Then result = False
            If Multiplex = True Then
                'Non-Essential RPC Bit Allocated
                result = False
                MessageBox.Show("RPC Address 3 is not required with Multiplexed Operation - Please Set to -1", mrcc)
            End If
        End If
        Return result
    End Function

    'Update Details
    Public Sub UpdateDetails(ByVal SigID As String, ByVal DefAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPCRef1 As Integer, ByVal RPCRef2 As Integer, ByVal RPCRef3 As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.UpdateSigDetails(SigID, DefAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPCRef1
        RPCBit2 = RPCRef2
        RPCBit3 = RPCRef3
    End Sub

    'OPERATE MODE FUNCTIONS

    Public Overridable Sub UpdateAspect() Implements ISignal.UpdateAspect
        Dim NextAspect As Integer = MyBase.GetNextSigAspect
        Select Case (NextAspect)
            Case -1 'No Route Set or Misc Conditions Failed
                'Change to Red
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Already Red
                    Case 1
                        'Yellow - Red
                        MyBase.ChangeSymbol(MyBase.Symbol - 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 3
                        'Green - Red
                        MyBase.ChangeSymbol(MyBase.Symbol - 3)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case Else
                        MessageBox.Show("Signal, " & MyBase.SignalID & " Invalid Aspect Code - " & MyBase.SigAspect, mrcc)
                End Select
                MyBase.SigAspect = 0
                If Multiplex = False Then
                    'Assume 1=Red,2=Yellow,3=Green
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                    If Not RPCBit3 = -1 Then HW.ClearBit(RPCBit3)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                End If
            Case 0 'Red
                'Change to Yellow
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Red - Yellow
                        MyBase.ChangeSymbol(MyBase.Symbol + 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 1
                        'Already Yellow
                    Case 3
                        'Green - Yellow
                        MyBase.ChangeSymbol(MyBase.Symbol - 2)
                        GUI.UpdateDisplay(MyBase.SignalID)
                End Select
                MyBase.SigAspect = 1
                If Multiplex = False Then
                    'Assume 1=Red,2=BottomYellow,3=Green
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                    If Not RPCBit3 = -1 Then HW.ClearBit(RPCBit3)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                End If

            Case Else 'Yellow, DYellow or Green
                'Change to Green
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Red - Green
                        MyBase.ChangeSymbol(MyBase.Symbol + 3)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 1
                        'Yellow - Green
                        MyBase.ChangeSymbol(MyBase.Symbol + 2)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 3
                        'Already Green
                    Case Else
                        MessageBox.Show("Signal, " & MyBase.SignalID & " Invalid Aspect Code - " & MyBase.SigAspect, mrcc)
                End Select
                MyBase.SigAspect = 3
                If Multiplex = False Then
                    'Assume 1=Red,2=BottomYellow,3=Green
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                    If Not RPCBit3 = -1 Then HW.SetBit(RPCBit3)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                End If
        End Select
    End Sub

    Public Overridable Sub EnterDesignMode() Implements ISignal.EnterDesignMode
        'Symbol to Grey
        Select Case (MyBase.SigAspect)
            Case 0
                'Red - Grey
                MyBase.ChangeSymbol(MyBase.Symbol - 2)
                GUI.UpdateDisplay(MyBase.SignalID)
            Case 1
                'Yellow - Grey
                MyBase.ChangeSymbol(MyBase.Symbol - 3)
                GUI.UpdateDisplay(MyBase.SignalID)
            Case 3
                'Green - Gray
                MyBase.ChangeSymbol(MyBase.Symbol - 5)
                GUI.UpdateDisplay(MyBase.SignalID)
        End Select
        'Aspect to Red
        MyBase.SigAspect = 0
    End Sub

    Public Overridable Sub EnterOperateMode() Implements ISignal.EnterOperateMode
        MyBase.SigAspect = MyBase.SigDefAspect
        MyBase.ChangeSymbol(MyBase.Symbol + MyBase.SigAspect + 2)
        GUI.UpdateDisplay(MyBase.SignalID)
        'HW set by first cycle of Update
    End Sub
End Class

'SIGNAL 4 ASPECT
'
<Serializable()> Public Class Signal4Aspect
    Inherits Signal
    Implements ISignal

    Private Multiplex As Boolean = False
    Private RPCBit1 As Integer = -1
    Private RPCBit2 As Integer = -1
    Private RPCBit3 As Integer = -1
    Private RPCBit4 As Integer = -1

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal ID As String, ByVal DefaultAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPC1 As Integer, ByVal RPC2 As Integer, ByVal RPC3 As Integer, ByVal RPC4 As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.New(x, y, ID, DefaultAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPC1
        RPCBit2 = RPC2
        RPCBit3 = RPC3
        RPCBit4 = RPC4
    End Sub


    Public ReadOnly Property SignalNumber() As String Implements ISignal.SigNumber
        Get
            Return MyBase.SignalID
        End Get
    End Property

    Public ReadOnly Property Multiplexed() As Boolean
        Get
            Return Multiplex
        End Get
    End Property

    Public ReadOnly Property RPC1() As Integer
        Get
            Return RPCBit1
        End Get
    End Property

    Public ReadOnly Property RPC2() As Integer
        Get
            Return RPCBit2
        End Get
    End Property

    Public ReadOnly Property RPC3() As Integer
        Get
            Return RPCBit3
        End Get
    End Property

    Public ReadOnly Property RPC4() As Integer
        Get
            Return RPCBit4
        End Get
    End Property

    'Validate
    Public Function validate() As Boolean
        Dim result As Boolean = True
        If MyBase.ValidateSig = False Then result = False

        'Check RPC Reference 1
        If Not RPCBit1 < 0 Then
            If Data.ValidateRPCBit(RPCBit1) = False Then result = False
        End If

        'Check RPC Reference 2
        If Not RPCBit2 < 0 Then
            If Data.ValidateRPCBit(RPCBit2) = False Then result = False
        End If

        'Check RPC Reference 3
        If Not RPCBit3 < 0 Then
            If Data.ValidateRPCBit(RPCBit3) = False Then result = False
            If Multiplex = True Then
                'Non-Essential RPC Bit Allocated
                result = False
                MessageBox.Show("RPC Address 3 is not required with Multiplexed Operation - Please Set to -1", mrcc)
            End If
        End If

        'Check RPC Reference 4
        If Not RPCBit4 < 0 Then
            If Data.ValidateRPCBit(RPCBit4) = False Then result = False
            If Multiplex = True Then
                'Non-Essential RPC Bit Allocated
                result = False
                MessageBox.Show("RPC Address 4 is not required with Multiplexed Operation - Please Set to -1", mrcc)
            End If
        End If
        Return result
    End Function

    'Update Details
    Public Sub UpdateDetails(ByVal SigID As String, ByVal DefAspect As Integer, ByVal SigSymbol As Integer, ByVal MultiTF As Boolean, ByVal RPCRef1 As Integer, ByVal RPCRef2 As Integer, ByVal RPCRef3 As Integer, ByVal RPCRef4 As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.UpdateSigDetails(SigID, DefAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
        Multiplex = MultiTF
        RPCBit1 = RPCRef1
        RPCBit2 = RPCRef2
        RPCBit3 = RPCRef3
        RPCBit4 = RPCRef4
    End Sub

    'OPERATE MODE FUNCTIONS

    Public Overridable Sub UpdateAspect() Implements ISignal.UpdateAspect
        Dim NextAspect As Integer = MyBase.GetNextSigAspect
        Select Case (NextAspect)
            Case -1 'No Route Set or Misc Conditions Failed
                'Change to Red
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Already Red
                    Case 1
                        'Yellow - Red
                        MyBase.ChangeSymbol(MyBase.Symbol - 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 2
                        'DYellow - Red
                        MyBase.ChangeSymbol(MyBase.Symbol - 2)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 3
                        'Green - Red
                        MyBase.ChangeSymbol(MyBase.Symbol - 3)
                        GUI.UpdateDisplay(MyBase.SignalID)
                End Select
                MyBase.SigAspect = 0
                If Multiplex = False Then
                    'Assume 1=Red,2=BottomYellow,3=Green,4=TopYellow
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                    If Not RPCBit3 = -1 Then HW.ClearBit(RPCBit3)
                    If Not RPCBit4 = -1 Then HW.ClearBit(RPCBit4)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                End If
            Case 0 'Red
                'Change to Yellow
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Red - Yellow
                        MyBase.ChangeSymbol(MyBase.Symbol + 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 1
                        'Already Yellow
                    Case 2
                        'DYellow - Yellow
                        MyBase.ChangeSymbol(MyBase.Symbol - 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 3
                        'Green - Yellow
                        MyBase.ChangeSymbol(MyBase.Symbol - 2)
                        GUI.UpdateDisplay(MyBase.SignalID)
                End Select
                MyBase.SigAspect = 1
                If Multiplex = False Then
                    'Assume 1=Red,2=BottomYellow,3=Green,4=TopYellow
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                    If Not RPCBit3 = -1 Then HW.ClearBit(RPCBit3)
                    If Not RPCBit4 = -1 Then HW.ClearBit(RPCBit4)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                End If
            Case 1 'Yellow
                'Change to DYellow
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Red - DYellow
                        MyBase.ChangeSymbol(MyBase.Symbol + 2)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 1
                        'Yellow - DYellow
                        MyBase.ChangeSymbol(MyBase.Symbol + 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 2
                        'Already DYellow
                    Case 3
                        'Green - DYellow
                        MyBase.ChangeSymbol(MyBase.Symbol - 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                End Select
                MyBase.SigAspect = 2
                If Multiplex = False Then
                    'Assume 1=Red,2=BottomYellow,3=Green,4=TopYellow
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                    If Not RPCBit3 = -1 Then HW.ClearBit(RPCBit3)
                    If Not RPCBit4 = -1 Then HW.SetBit(RPCBit4)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.SetBit(RPCBit2)
                End If
            Case Else 'DYellow or Green
                'Change to Green
                Select Case (MyBase.SigAspect)
                    Case 0
                        'Red - Green
                        MyBase.ChangeSymbol(MyBase.Symbol + 3)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 1
                        'Yellow - Green
                        MyBase.ChangeSymbol(MyBase.Symbol + 2)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 2
                        'DYellow - Green
                        MyBase.ChangeSymbol(MyBase.Symbol + 1)
                        GUI.UpdateDisplay(MyBase.SignalID)
                    Case 3
                        'Already Green
                End Select
                MyBase.SigAspect = 3
                If Multiplex = False Then
                    'Assume 1=Red,2=BottomYellow,3=Green,4=TopYellow
                    If Not RPCBit1 = -1 Then HW.ClearBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                    If Not RPCBit3 = -1 Then HW.SetBit(RPCBit3)
                    If Not RPCBit4 = -1 Then HW.ClearBit(RPCBit4)
                Else
                    'Assume Wired to MERG SD4 Spec
                    If Not RPCBit1 = -1 Then HW.SetBit(RPCBit1)
                    If Not RPCBit2 = -1 Then HW.ClearBit(RPCBit2)
                End If
        End Select
    End Sub

    Public Overridable Sub EnterDesignMode() Implements ISignal.EnterDesignMode
        'Symbol to DGrey
        Select Case (MyBase.SigAspect)
            Case 0
                'Red - DGrey
                MyBase.ChangeSymbol(MyBase.Symbol - 1)
                GUI.UpdateDisplay(MyBase.SignalID)
            Case 1
                'Yellow - DGrey
                MyBase.ChangeSymbol(MyBase.Symbol - 2)
                GUI.UpdateDisplay(MyBase.SignalID)
            Case 2
                'DYellow - DGrey
                MyBase.ChangeSymbol(MyBase.Symbol - 3)
                GUI.UpdateDisplay(MyBase.SignalID)
            Case 3
                'Green - DGray
                MyBase.ChangeSymbol(MyBase.Symbol - 4)
                GUI.UpdateDisplay(MyBase.SignalID)
        End Select
        'Aspect to Red
        MyBase.SigAspect = 0
    End Sub

    Public Overridable Sub EnterOperateMode() Implements ISignal.EnterOperateMode
        MyBase.SigAspect = MyBase.SigDefAspect
        MyBase.ChangeSymbol(MyBase.Symbol + MyBase.SigAspect + 1)
        GUI.UpdateDisplay(MyBase.SignalID)
        'HW set by first cycle of Update
    End Sub

End Class

'ISOLATED EXIT
'
<Serializable()> Public Class IsolatedExit
    Inherits Signal
    Implements ISignal

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal ID As String, ByVal DefaultAspect As Integer, ByVal SigSymbol As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.New(x, y, ID, DefaultAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
    End Sub


    Public ReadOnly Property SignalNumber() As String Implements ISignal.SigNumber
        Get
            Return MyBase.SignalID
        End Get
    End Property

    'Validate
    Public Function validate() As Boolean
        Return MyBase.ValidateSig()
    End Function

    'Update Details
    Public Sub UpdateDetails(ByVal SigID As String, ByVal DefAspect As Integer, ByVal SigSymbol As Integer, ByVal Routes As ArrayList, ByVal ILMiscTrue As String, ByVal ILMiscFalse As String, ByVal ILAutoSig As Boolean)
        MyBase.UpdateSigDetails(SigID, DefAspect, SigSymbol, Routes, ILMiscTrue, ILMiscFalse, ILAutoSig)
    End Sub

    'OPERATE MODE FUNCTIONS

    Public Overridable Sub UpdateAspect() Implements ISignal.UpdateAspect
        'Aspect of Isolated Exit doesn't Change
    End Sub

    Public Overridable Sub EnterDesignMode() Implements ISignal.EnterDesignMode
        'Aspect to Red
        MyBase.SigAspect = 0
    End Sub

    Public Overridable Sub EnterOperateMode() Implements ISignal.EnterOperateMode
        MyBase.SigAspect = MyBase.SigDefAspect
    End Sub
End Class

'MISC DEVICES
'************
<Serializable()> Public Class StaticMisc
    Inherits ScreenElements
    Implements IMisc

    Private ID As String = ""
    Private State As Boolean = False
    Private Type As String = "Hidden"
    Private MiscTypeCode As Integer 'This represents Type and Colour to Help Form
    Private SymbolCode As Integer = 133
    Private RPCBit As Integer = -1

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal DeviceID As String, ByVal SelectedType As Integer, ByVal RPCRef As Integer)
        MyBase.New(x, y)
        ID = DeviceID
        MiscTypeCode = SelectedType
        RPCBit = RPCRef

        Select Case (SelectedType)
            Case 0 'Switch
                SymbolCode = 133
                Type = "Switch"
            Case 1 'Control(Grey)
                SymbolCode = 135
                Type = "Control"
            Case 2 'Control(Green)
                SymbolCode = 136
                Type = "Control"
            Case 3 'Control(Red)
                SymbolCode = 137
                Type = "Control"
            Case 4 'Control(Yellow)
                SymbolCode = 138
                Type = "Control"
            Case 5 'LED(Grey)
                SymbolCode = 139
                Type = "LED"
            Case 6 'LED(Green)
                SymbolCode = 140
                Type = "LED"
            Case 7 'LED(Red)
                SymbolCode = 141
                Type = "LED"
            Case 8 'LED(Yellow)
                SymbolCode = 142
                Type = "LED"
            Case 9 'Auto Button
                SymbolCode = 146
                Type = "Auto"
            Case 10 'Hidden
                SymbolCode = 135
                Type = "Hidden"
        End Select
    End Sub

    Friend Sub UpdateDetails(ByVal DeviceID As String, ByVal SelectedType As Integer, ByVal RPCRef As Integer)
        'Validated by Form
        ID = DeviceID
        MiscTypeCode = SelectedType
        RPCBit = RPCRef

        Select Case (SelectedType)
            Case 0 'Switch
                SymbolCode = 133
                Type = "Switch"
            Case 1 'Control(Grey)
                SymbolCode = 135
                Type = "Control"
            Case 2 'Control(Green)
                SymbolCode = 136
                Type = "Control"
            Case 3 'Control(Red)
                SymbolCode = 137
                Type = "Control"
            Case 4 'Control(Yellow)
                SymbolCode = 138
                Type = "Control"
            Case 5 'LED(Grey)
                SymbolCode = 139
                Type = "LED"
            Case 6 'LED(Green)
                SymbolCode = 140
                Type = "LED"
            Case 7 'LED(Red)
                SymbolCode = 141
                Type = "LED"
            Case 8 'LED(Yellow)
                SymbolCode = 142
                Type = "LED"
            Case 9 'Auto Button
                SymbolCode = 146
                Type = "Auto"
            Case 10 'Hidden
                SymbolCode = 135
                Type = "Hidden"
        End Select
    End Sub

    Public Sub ChangeSymbol(ByVal symbol As Integer)
        SymbolCode = symbol
    End Sub

    Public Function Validate() As Boolean
        Dim result As Boolean = True

        ' Check ID
        If DeviceID = "" Then
            result = False
            MessageBox.Show("No Device ID given - Please Enter One", mrcc)
        Else
            If Data.CheckReference(DeviceID) <> -1 Then '-1 means not found
                'Reference in use already
                Dim p As Point
                p = CType(Data.NameIndex(DeviceID), Point)
                If p.X <> CurrentX Or p.Y <> CurrentY Then 'checks if match is itself
                    'Name Exists and in use by another object
                    result = False
                    MessageBox.Show("Device ID " & DeviceID & " is already in use - Please Choose Another", mrcc)
                Else
                    'Name Exists for this object
                End If
            Else
                'Reference not in use

            End If

            'Check format is "SigNo Auto" if Auto Button
            If Type = "Auto" Then
                If Len(ID) < 5 Then
                    result = False
                    MessageBox.Show("Device ID " & DeviceID & " : Auto Button IDs must be in the form 'SigNo Auto'", mrcc)
                Else
                    If Right(ID, 5) <> " Auto" Then
                        result = False
                        MessageBox.Show("Device ID " & DeviceID & " : Auto Button IDs must be in the form 'SigNo Auto'", mrcc)
                        If Data.CheckReference(Left(ID, Len(ID) - Len(" Auto"))) > 6 Or Data.CheckReference(Left(ID, Len(ID) - Len(" Auto"))) = -1 Then
                            result = False
                            MessageBox.Show("Device ID " & DeviceID & " : Auto Button's Referenced Signal ID doesn't exist as a Signal", mrcc)
                        End If
                    End If
                End If
            End If
        End If
        'Check RPC Reference
        If RPCBit < 0 Then
            If Me.MiscTypeCode = 10 Then
                result = False
                MessageBox.Show("Hidden Misc Devices MUST have an RPC Reference", mrcc)
            Else
                If Me.MiscTypeCode > 4 And Me.MiscTypeCode < 9 Then
                    result = False
                    MessageBox.Show("Input LEDs MUST have an RPC Reference", mrcc)
                End If
            End If
        Else
            If Me.MiscTypeCode = 9 Then
                result = False
                MessageBox.Show("Auto Buttons are NOT allowed to have an RPC Reference", mrcc)
            Else
                If Data.ValidateRPCBit(RPCBit) = False Then result = False
            End If
        End If

        Return result
    End Function

    Public Property CurrentState() As Boolean Implements IMisc.State
        Get
            Return State
        End Get

        Set(ByVal Value As Boolean)
            If Value = True And State = False Then
                'Set Misc
                Select Case (Type)
                    Case "Switch", "Auto"
                        ChangeSymbol(SymbolCode + 1)
                        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                    Case "Control", "LED", "Hidden"
                        If Not (Type = "Hidden" And Mode = 2) Then
                            'Executes if Not Hidden in Operate Mode
                            ChangeSymbol(SymbolCode + 4)
                            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                        End If
                    Case Else
                End Select
                If (Type = "Switch" Or Type = "Control") And RPCBit <> -1 Then HW.SetBit(RPCBit)
            Else
                If Value = False And State = True Then
                    'Clear Misc
                    Select Case (Type)
                        Case "Switch", "Auto"
                            ChangeSymbol(SymbolCode - 1)
                            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                        Case "Control", "LED", "Hidden"
                            If Not (Type = "Hidden" And Mode = 2) Then
                                'Executes if Not Hidden in Operate Mode
                                ChangeSymbol(SymbolCode - 4)
                                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                            End If
                        Case Else
                    End Select
                End If
                If (Type = "Switch" Or Type = "Control") And RPCBit <> -1 Then HW.ClearBit(RPCBit)
            End If
            State = Value
        End Set
    End Property

    Public ReadOnly Property Symbol() As Integer
        Get
            Return SymbolCode
        End Get
    End Property

    Public ReadOnly Property MiscType() As String
        Get
            Return Type
        End Get
    End Property

    Public ReadOnly Property TypeCode() As Integer
        Get
            Return MiscTypeCode
        End Get
    End Property

    Public ReadOnly Property DeviceID() As String Implements IMisc.ID
        Get
            Return ID
        End Get
    End Property

    Public ReadOnly Property RPCRef() As Integer
        Get
            Return RPCBit
        End Get
    End Property

    'OPERATE MODE FUNCTIONS
    Public Sub Update() Implements IMisc.Update
        'Updates State and Symbol as required
        If Type = "Hidden" Or Type = "LED" Then
            'Read state from RPC
            If RPCBit <> -1 Then
                If HW.ReadBit(RPCBit) = 1 Then
                    CurrentState = False 'Was True
                Else
                    CurrentState = True 'Was False
                End If
            End If
        End If
    End Sub

    Public Function SetMisc() As String
        Dim FeedbackText As String = ""
        'Handles Mouse Click on Static Misc
        If Not (Type = "Hidden" Or Type = "LED") Then
            If Type <> "Auto" Then
                CurrentState = True
            Else
                'This is a Request to Set an A Button
                'First Check a Route is Set from the Entrance Signal
                'Assume that ID is "SigNo Auto" - Read "SigNo" as all before Space - This is Validated

                Dim SigNo As String = ""
                SigNo = Left(ID, Len(ID) - Len(" Auto"))
                If Data.IsRouteSetFromSig(SigNo) = True Then
                    CurrentState = True
                    'Route Set - A button gets set.
                Else
                    'Route Not Set from Sig
                    FeedbackText = "Unable to Set " & ID & " : Route must be set first"
                End If
            End If
        Else
            If Mode = 1 And RPCBit <> -1 Then
                'Simulates Input if Test Mode
                HW.SimulateInput(RPCBit, False)
            End If
        End If
        Return FeedbackText
    End Function

    Public Sub CancelMisc()
        If Not (Type = "Hidden" Or Type = "LED") Then
            CurrentState = False
        Else
            If Mode = 1 And RPCBit <> -1 Then
                'Simulates Input if Test Mode
                HW.SimulateInput(RPCBit, True)
            End If
        End If
    End Sub

    Public Sub EnterDesignMode() Implements IMisc.EnterDesignMode
        CurrentState = False

        If Type = "LED" Then
            'Change to Set Symbol to distinguish from Control Device
            ChangeSymbol(SymbolCode + 4)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If

        If Type = "Hidden" Then
            'Make Viewable
            ChangeSymbol(135)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If
    End Sub

    Public Sub EnterTestMode() Implements IMisc.EnterTestMode
        If Mode = 0 Then
            'Changing from Design - Test
            If Type = "LED" Then
                'Change back to Symbol for State:False
                ChangeSymbol(SymbolCode - 4)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
            CurrentState = False
        Else
            'Changing from Operate - Test
            CurrentState = False
            If Type = "Hidden" Then
                ChangeSymbol(135)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If
    End Sub

    Public Sub EnterOperateMode() Implements IMisc.EnterOperateMode
        If Mode = 0 Then
            'Changing from Design - Operate
            If Type = "LED" Then
                'Change back to Symbol for State:False
                ChangeSymbol(SymbolCode - 4)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If
        CurrentState = False
        If Type = "Hidden" Then
            'Hide Hidden device
            ChangeSymbol(156)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If
    End Sub
End Class

<Serializable()> Public Class MiscOutput
    Inherits ScreenElements
    Implements IMisc

    Private ID As String = ""
    Private State As Boolean = False
    Private Type As String = "Hidden"
    Private MiscTypeCode As Integer 'This represents Type and Colour to Help Form
    Private SymbolCode As Integer = 139
    Private RPCBit As Integer = -1

    Private AndTF As Boolean = True

    Private ILActive1 As Boolean = False
    Private ILDeviceType1 As Integer = -1
    Private ILDeviceID1 As String = ""
    Private ILCondition1 As String = ""

    Private ILActive2 As Boolean = False
    Private ILDeviceType2 As Integer = -1
    Private ILDeviceID2 As String = ""
    Private ILCondition2 As String = ""

    Private ILActive3 As Boolean = False
    Private ILDeviceType3 As Integer = -1
    Private ILDeviceID3 As String = ""
    Private ILCondition3 As String = ""

    Private ILActive4 As Boolean = False
    Private ILDeviceType4 As Integer = -1
    Private ILDeviceID4 As String = ""
    Private ILCondition4 As String = ""

    Private ILActive5 As Boolean = False
    Private ILDeviceType5 As Integer = -1
    Private ILDeviceID5 As String = ""
    Private ILCondition5 As String = ""

    'Constructor
    Sub New(ByVal x As Integer, ByVal y As Integer, ByVal DeviceID As String, ByVal SelectedType As Integer, ByVal RPCRef As Integer, ByVal ILAndTF As Boolean, ByVal ILActiveChk1 As Boolean, ByVal ILDeviceTypes1 As Integer, ByVal ILDeviceIDs1 As String, ByVal ILConditions1 As String, ByVal ILActiveChk2 As Boolean, ByVal ILDeviceTypes2 As Integer, ByVal ILDeviceIDs2 As String, ByVal ILConditions2 As String, ByVal ILActiveChk3 As Boolean, ByVal ILDeviceTypes3 As Integer, ByVal ILDeviceIDs3 As String, ByVal ILConditions3 As String, ByVal ILActiveChk4 As Boolean, ByVal ILDeviceTypes4 As Integer, ByVal ILDeviceIDs4 As String, ByVal ILConditions4 As String, ByVal ILActiveChk5 As Boolean, ByVal ILDeviceTypes5 As Integer, ByVal ILDeviceIDs5 As String, ByVal ILConditions5 As String)
        MyBase.New(x, y)
        ID = DeviceID
        MiscTypeCode = SelectedType
        RPCBit = RPCRef

        Select Case (SelectedType)
            Case 0 'LED(Grey)
                SymbolCode = 139
                Type = "LED"
            Case 1 'LED(Green)
                SymbolCode = 140
                Type = "LED"
            Case 2 'LED(Red)
                SymbolCode = 141
                Type = "LED"
            Case 3 'LED(Yellow)
                SymbolCode = 142
                Type = "LED"
            Case 4 'Hidden
                SymbolCode = 135
                Type = "Hidden"
        End Select

        AndTF = ILAndTF
        ILActive1 = ILActiveChk1
        ILDeviceType1 = ILDeviceTypes1
        ILDeviceID1 = ILDeviceIDs1
        ILCondition1 = ILConditions1
        ILActive2 = ILActiveChk2
        ILDeviceType2 = ILDeviceTypes2
        ILDeviceID2 = ILDeviceIDs2
        ILCondition2 = ILConditions2
        ILActive3 = ILActiveChk3
        ILDeviceType3 = ILDeviceTypes3
        ILDeviceID3 = ILDeviceIDs3
        ILCondition3 = ILConditions3
        ILActive4 = ILActiveChk4
        ILDeviceType4 = ILDeviceTypes4
        ILDeviceID4 = ILDeviceIDs4
        ILCondition4 = ILConditions4
        ILActive5 = ILActiveChk5
        ILDeviceType5 = ILDeviceTypes5
        ILDeviceID5 = ILDeviceIDs5
        ILCondition5 = ILConditions5
    End Sub

    Friend Sub UpdateDetails(ByVal DeviceID As String, ByVal SelectedType As Integer, ByVal RPCRef As Integer, ByVal ILAndTF As Boolean, ByVal ILActiveChk1 As Boolean, ByVal ILDeviceTypes1 As Integer, ByVal ILDeviceIDs1 As String, ByVal ILConditions1 As String, ByVal ILActiveChk2 As Boolean, ByVal ILDeviceTypes2 As Integer, ByVal ILDeviceIDs2 As String, ByVal ILConditions2 As String, ByVal ILActiveChk3 As Boolean, ByVal ILDeviceTypes3 As Integer, ByVal ILDeviceIDs3 As String, ByVal ILConditions3 As String, ByVal ILActiveChk4 As Boolean, ByVal ILDeviceTypes4 As Integer, ByVal ILDeviceIDs4 As String, ByVal ILConditions4 As String, ByVal ILActiveChk5 As Boolean, ByVal ILDeviceTypes5 As Integer, ByVal ILDeviceIDs5 As String, ByVal ILConditions5 As String)
        ID = DeviceID
        MiscTypeCode = SelectedType
        RPCBit = RPCRef

        Select Case (SelectedType)
            Case 0 'LED(Grey)
                SymbolCode = 139
                Type = "LED"
            Case 1 'LED(Green)
                SymbolCode = 140
                Type = "LED"
            Case 2 'LED(Red)
                SymbolCode = 141
                Type = "LED"
            Case 3 'LED(Yellow)
                SymbolCode = 142
                Type = "LED"
            Case 4 'Hidden
                SymbolCode = 135
                Type = "Hidden"
        End Select

        AndTF = ILAndTF
        ILActive1 = ILActiveChk1
        ILDeviceType1 = ILDeviceTypes1
        ILDeviceID1 = ILDeviceIDs1
        ILCondition1 = ILConditions1
        ILActive2 = ILActiveChk2
        ILDeviceType2 = ILDeviceTypes2
        ILDeviceID2 = ILDeviceIDs2
        ILCondition2 = ILConditions2
        ILActive3 = ILActiveChk3
        ILDeviceType3 = ILDeviceTypes3
        ILDeviceID3 = ILDeviceIDs3
        ILCondition3 = ILConditions3
        ILActive4 = ILActiveChk4
        ILDeviceType4 = ILDeviceTypes4
        ILDeviceID4 = ILDeviceIDs4
        ILCondition4 = ILConditions4
        ILActive5 = ILActiveChk5
        ILDeviceType5 = ILDeviceTypes5
        ILDeviceID5 = ILDeviceIDs5
        ILCondition5 = ILConditions5
    End Sub

    Public Sub ChangeSymbol(ByVal symbol As Integer)
        SymbolCode = symbol
    End Sub

    Public Function Validate() As Boolean
        Dim result As Boolean = True

        ' Check ID
        If DeviceID = "" Then
            result = False
            MessageBox.Show("No Device ID given - Please Enter One", mrcc)
        Else
            If Data.CheckReference(DeviceID) <> -1 Then '-1 means not found
                'Reference in use already
                Dim p As Point
                p = CType(Data.NameIndex(DeviceID), Point)
                If p.X <> CurrentX Or p.Y <> CurrentY Then 'checks if match is itself
                    'Name Exists and in use by another object
                    result = False
                    MessageBox.Show("Device ID " & DeviceID & " is already in use - Please Choose Another", mrcc)
                Else
                    'Name Exists for this object
                End If
            Else
                'Reference not in use
            End If
        End If

        'Check RPC Reference
        If Not RPCBit = -1 Then
            If Data.ValidateRPCBit(RPCBit) = False Then result = False
        End If

        'Check Active Conditions have Type, ID and Condition entered
        If ILActive1 = True And (ILDeviceType1 = -1 Or ILCondition1 = "" Or ILDeviceID1 = "") Then
            result = False
            MessageBox.Show("Line 1: Active Conditions must have data entered", mrcc)
        End If
        If ILActive2 = True And (ILDeviceType2 = -1 Or ILCondition2 = "" Or ILDeviceID2 = "") Then
            result = False
            MessageBox.Show("Line 2: Active Conditions must have data entered", mrcc)
        End If
        If ILActive3 = True And (ILDeviceType3 = -1 Or ILCondition3 = "" Or ILDeviceID3 = "") Then
            result = False
            MessageBox.Show("Line 3: Active Conditions must have data entered", mrcc)
        End If
        If ILActive4 = True And (ILDeviceType4 = -1 Or ILCondition4 = "" Or ILDeviceID4 = "") Then
            result = False
            MessageBox.Show("Line 4: Active Conditions must have data entered", mrcc)
        End If
        If ILActive5 = True And (ILDeviceType5 = -1 Or ILCondition5 = "" Or ILDeviceID5 = "") Then
            result = False
            MessageBox.Show("Line 5: Active Conditions must have data entered", mrcc)
        End If

        'Validate Conditions to Device Types
        If Not CheckConditionValid(ILDeviceType1, ILCondition1) Then result = False
        If Not CheckConditionValid(ILDeviceType2, ILCondition2) Then result = False
        If Not CheckConditionValid(ILDeviceType3, ILCondition3) Then result = False
        If Not CheckConditionValid(ILDeviceType4, ILCondition4) Then result = False
        If Not CheckConditionValid(ILDeviceType5, ILCondition5) Then result = False

        Return result
    End Function

    Private Function CheckConditionValid(ByVal Type As Integer, ByVal Condition As String) As Boolean
        Dim result As Boolean = True
        If Type <> -1 And Condition <> "" Then
            Select Case (Type)
                Case 2
                    If Not (Condition = "Not Red" Or Condition = "Red" Or Condition = "Green") Then result = False
                Case 4
                    If Not (Condition = "Not Red" Or Condition = "Red" Or Condition = "Yellow" Or Condition = "Green") Then result = False
                Case 6
                    If Not (Condition = "Not Red" Or Condition = "Red" Or Condition = "Yellow" Or Condition = "Double Yellow" Or Condition = "Green") Then result = False
                Case 0
                    If Not (Condition = "Not Red" Or Condition = "Stop" Or Condition = "Proceed") Then result = False
                Case 15
                    If Not (Condition = "Set" Or Condition = "Not Set" Or Condition = "Tracks Clear" Or Condition = "Tracks Occupied") Then result = False
                Case 10
                    If Not (Condition = "Normal" Or Condition = "Reverse") Then result = False
                Case 11
                    If Not (Condition = "Occupied" Or Condition = "Clear") Then result = False
                Case 12
                    If Not (Condition = "True" Or Condition = "False") Then result = False
                Case Else
                    MessageBox.Show("Error - Invalid Type", mrcc)
                    result = False
            End Select
        End If
        If result = False Then MessageBox.Show("Error - " & GetDeviceTypeStr(Type) & " can not be " & Condition, mrcc)
        Return result
    End Function

    Public Function ValidateILData() As String
        Dim errorreport As String = ""
        errorreport &= CheckLineValid(ILActive1, ILDeviceType1, ILDeviceID1)
        errorreport &= CheckLineValid(ILActive2, ILDeviceType2, ILDeviceID2)
        errorreport &= CheckLineValid(ILActive3, ILDeviceType3, ILDeviceID3)
        errorreport &= CheckLineValid(ILActive4, ILDeviceType4, ILDeviceID4)
        errorreport &= CheckLineValid(ILActive5, ILDeviceType5, ILDeviceID5)

        Return errorreport
    End Function

    Private Function CheckLineValid(ByVal CheckTF As Boolean, ByVal DevType As Integer, ByVal DeviceID As String) As String
        Dim errorreport As String = ""
        If CheckTF Then
            If DevType = 15 Then
                If Not Data.CheckRouteExists(DeviceID) Then errorreport &= "Misc Output " & ID & ", Route " & DeviceID & " Does Not Exist" & vbCrLf
            Else
                Dim TempType As Integer = Data.CheckReference(DeviceID)
                If Not (DevType = TempType Or (TempType = 14 And DevType = 12)) Then errorreport &= "Misc Output " & ID & ", " & GetDeviceTypeStr(DevType) & " " & DeviceID & " Does Not Exist" & vbCrLf
            End If
        End If
        Return errorreport
    End Function

    Private Function GetDeviceTypeStr(ByVal DeviceType As Integer) As String
        Dim str As String = ""
        Select Case (DeviceType)
            Case 0
                str = "Shunt Signal"
            Case 2
                str = "2 Aspect Signal"
            Case 4
                str = "3 Aspect Signal"
            Case 6
                str = "4 Aspect Signal"
            Case 10
                str = "Point"
            Case 11
                str = "Track Section"
            Case 12
                str = "Misc Device"
            Case 15
                str = "Route"
        End Select
        Return str
    End Function

    Public Property CurrentState() As Boolean Implements IMisc.State

        Get
            Return State
        End Get
        Set(ByVal Value As Boolean)
            If Value = True And State = False Then
                'Output to True
                If Not (Mode = 2 And Type = "Hidden") Then
                    ChangeSymbol(SymbolCode + 4)
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                Else
                    'Hidden Device in Operate Mode - Do nothing
                End If
                If RPCBit <> -1 Then HW.SetBit(RPCBit)
            Else
                If Value = False And State = True Then
                    'Output to False
                    If Not (Mode = 2 And Type = "Hidden") Then
                        ChangeSymbol(SymbolCode - 4)
                        GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                    Else
                        'Hidden Device in Operate Mode - Do nothing
                    End If
                    If RPCBit <> -1 Then HW.ClearBit(RPCBit)
                End If
            End If
            State = Value
        End Set
    End Property

    Public ReadOnly Property Symbol() As Integer
        Get
            Return SymbolCode
        End Get
    End Property

    Public ReadOnly Property MiscType() As String
        Get
            Return Type
        End Get
    End Property

    Public ReadOnly Property TypeCode() As Integer
        Get
            Return MiscTypeCode
        End Get
    End Property

    Public ReadOnly Property DeviceID() As String Implements IMisc.ID
        Get
            Return ID
        End Get
    End Property

    Public ReadOnly Property RPCRef() As Integer
        Get
            Return RPCBit
        End Get
    End Property

    Public ReadOnly Property ILAndTF() As Boolean
        Get
            Return AndTF
        End Get
    End Property

    Public ReadOnly Property ILActiveChks1() As Boolean
        Get
            Return ILActive1
        End Get
    End Property

    Public ReadOnly Property ILDeviceTypes1() As Integer
        Get
            Return ILDeviceType1
        End Get
    End Property

    Public ReadOnly Property ILDeviceIDs1() As String
        Get
            Return ILDeviceID1
        End Get
    End Property

    Public ReadOnly Property ILConditions1() As String
        Get
            Return ILCondition1
        End Get
    End Property

    Public ReadOnly Property ILActiveChks2() As Boolean
        Get
            Return ILActive2
        End Get
    End Property

    Public ReadOnly Property ILDeviceTypes2() As Integer
        Get
            Return ILDeviceType2
        End Get
    End Property

    Public ReadOnly Property ILDeviceIDs2() As String
        Get
            Return ILDeviceID2
        End Get
    End Property

    Public ReadOnly Property ILConditions2() As String
        Get
            Return ILCondition2
        End Get
    End Property

    Public ReadOnly Property ILActiveChks3() As Boolean
        Get
            Return ILActive3
        End Get
    End Property

    Public ReadOnly Property ILDeviceTypes3() As Integer
        Get
            Return ILDeviceType3
        End Get
    End Property

    Public ReadOnly Property ILDeviceIDs3() As String
        Get
            Return ILDeviceID3
        End Get
    End Property

    Public ReadOnly Property ILConditions3() As String
        Get
            Return ILCondition3
        End Get
    End Property

    Public ReadOnly Property ILActiveChks4() As Boolean
        Get
            Return ILActive4
        End Get
    End Property

    Public ReadOnly Property ILDeviceTypes4() As Integer
        Get
            Return ILDeviceType4
        End Get
    End Property

    Public ReadOnly Property ILDeviceIDs4() As String
        Get
            Return ILDeviceID4
        End Get
    End Property

    Public ReadOnly Property ILConditions4() As String
        Get
            Return ILCondition4
        End Get
    End Property

    Public ReadOnly Property ILActiveChks5() As Boolean
        Get
            Return ILActive5
        End Get
    End Property

    Public ReadOnly Property ILDeviceTypes5() As Integer
        Get
            Return ILDeviceType5
        End Get
    End Property

    Public ReadOnly Property ILDeviceIDs5() As String
        Get
            Return ILDeviceID5
        End Get
    End Property

    Public ReadOnly Property ILConditions5() As String
        Get
            Return ILCondition5
        End Get
    End Property

    'OPERATE MODE FUNCTIONS
    Public Sub Update() Implements IMisc.Update
        'Updates State as required
        Dim result As Boolean = False

        If AndTF Then
            'AND
            result = True
            If ILActive1 = True Then result = result And EvaluateCondition(ILDeviceType1, ILDeviceID1, ILCondition1)
            If ILActive2 = True Then result = result And EvaluateCondition(ILDeviceType2, ILDeviceID2, ILCondition2)
            If ILActive3 = True Then result = result And EvaluateCondition(ILDeviceType3, ILDeviceID3, ILCondition3)
            If ILActive4 = True Then result = result And EvaluateCondition(ILDeviceType4, ILDeviceID4, ILCondition4)
            If ILActive5 = True Then result = result And EvaluateCondition(ILDeviceType5, ILDeviceID5, ILCondition5)
        Else
            'OR
            If ILActive1 = True Then result = result Or EvaluateCondition(ILDeviceType1, ILDeviceID1, ILCondition1)
            If ILActive2 = True Then result = result Or EvaluateCondition(ILDeviceType2, ILDeviceID2, ILCondition2)
            If ILActive3 = True Then result = result Or EvaluateCondition(ILDeviceType3, ILDeviceID3, ILCondition3)
            If ILActive4 = True Then result = result Or EvaluateCondition(ILDeviceType4, ILDeviceID4, ILCondition4)
            If ILActive5 = True Then result = result Or EvaluateCondition(ILDeviceType5, ILDeviceID5, ILCondition5)
        End If

        CurrentState = result
    End Sub

    Private Function EvaluateCondition(ByVal DevType As Integer, ByVal DevID As String, ByVal LineCond As String) As Boolean
        Dim result As Boolean = False
        Dim SigAsp As Integer = 0
        Dim TempStr As String = ""
        Select Case DevType
            Case 0 'Shunt Sig
                SigAsp = CType(Data.GetByName(DevID), SignalShunt).SigAspect
                Select Case LineCond
                    Case "Not Red"
                        If SigAsp <> 0 Then result = True
                    Case "Stop"
                        If SigAsp = 0 Then result = True
                    Case "Proceed"
                        If SigAsp = 1 Then result = True
                End Select
            Case 2 '2 Aspect
                SigAsp = CType(Data.GetByName(DevID), Signal2Aspect).SigAspect
                Select Case LineCond
                    Case "Not Red"
                        If SigAsp <> 0 Then result = True
                    Case "Red"
                        If SigAsp = 0 Then result = True
                    Case "Green"
                        If SigAsp = 3 Then result = True
                End Select
            Case 4 '3 Aspect
                SigAsp = CType(Data.GetByName(DevID), Signal3Aspect).SigAspect
                Select Case LineCond
                    Case "Not Red"
                        If SigAsp <> 0 Then result = True
                    Case "Red"
                        If SigAsp = 0 Then result = True
                    Case "Yellow"
                        If SigAsp = 1 Then result = True
                    Case "Green"
                        If SigAsp = 3 Then result = True
                End Select
            Case 6 '4 Aspect
                SigAsp = CType(Data.GetByName(DevID), Signal4Aspect).SigAspect
                Select Case LineCond
                    Case "Not Red"
                        If SigAsp <> 0 Then result = True
                    Case "Red"
                        If SigAsp = 0 Then result = True
                    Case "Yellow"
                        If SigAsp = 1 Then result = True
                    Case "Double Yellow"
                        If SigAsp = 2 Then result = True
                    Case "Green"
                        If SigAsp = 3 Then result = True
                End Select
            Case 10 'Point
                If Left(LineCond, 1) = Data.GetPoint(DevID).Position Then result = True
            Case 11 'Track Section
                result = Data.GetTKSection(DevID).TOcc
                If LineCond = "Clear" Then result = Not result
            Case 12 'Misc
                result = Data.GetMiscState(DevID)
                If LineCond = "False" Then result = Not result
            Case 15 'Route
                Select Case (LineCond)
                    Case "Set"
                        result = Data.GetRoute(DevID).USet
                    Case "Not Set"
                        result = Not Data.GetRoute(DevID).USet
                    Case "Tracks Clear"
                        result = Data.GetRoute(DevID).TracksClear
                    Case "Tracks Occupied"
                        result = Not Data.GetRoute(DevID).TracksClear
                End Select
        End Select

        Return result
    End Function

    Public Sub EnterDesignMode() Implements IMisc.EnterDesignMode
        CurrentState = False

        If Type = "LED" Then
            'Change Symbol to Set to distinguish from Control
            ChangeSymbol(SymbolCode + 4)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If

        If Type = "Hidden" Then
            'Show Hidden Device
            ChangeSymbol(135)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If
    End Sub

    Public Sub EnterTestMode() Implements IMisc.EnterTestMode
        If Mode = 0 Then
            'Design - Test
            If Type = "LED" Then
                'Return Symbol to Unset
                ChangeSymbol(SymbolCode - 4)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If

        CurrentState = False

        If Mode = 2 Then
            'Operate - Test
            If Type = "Hidden" Then
                ChangeSymbol(135)
                GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
            End If
        End If
    End Sub

    Public Sub EnterOperateMode() Implements IMisc.EnterOperateMode
        CurrentState = False

        If Type = "LED" And Mode = 0 Then
            'Return Symbol to Unset
            ChangeSymbol(SymbolCode - 4)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If

        If Type = "Hidden" Then
            'Show Hidden Device
            ChangeSymbol(156)
            GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
        End If
    End Sub

End Class

'ROUTES
'******
<Serializable()> Public Class Route
    Private RouteName As String = ""
    Private EntrySignal As String = ""
    Private ExitSignal As String = ""
    Private ConflictingRoutes As New ArrayList(10)
    Private PointsNormal As New ArrayList(10)
    Private PointsReverse As New ArrayList(10)
    Private TrackSections As New ArrayList(10)
    Private ILMiscTrue As String = ""
    Private ILMiscFalse As String = ""
    Private EarlyReleaseTk As String = ""
    Private EarlyReleaseClrOcc As String = "Occ"
    Private AButton As String = ""

    Private RouteSet As Boolean = False
    Private TracksClr As Boolean = False
    Private APressed As Boolean = False
    Private ERTkOcc As Boolean = False

    Sub New(ByVal ID As String, ByVal SigEntry As String, ByVal SigExit As String, ByVal Conflicting As ArrayList, ByVal NPoints As ArrayList, ByVal RPoints As ArrayList, ByVal Tracks As ArrayList, ByVal MiscTrue As String, ByVal MiscFalse As String, ByVal EarlyRelease As String, ByVal EarlyClrOcc As String, ByVal AButtonRef As String)
        RouteName = ID
        EntrySignal = SigEntry
        ExitSignal = SigExit
        ConflictingRoutes = Conflicting
        PointsNormal = NPoints
        PointsReverse = RPoints
        TrackSections = Tracks
        ILMiscTrue = MiscTrue
        ILMiscFalse = MiscFalse
        EarlyReleaseTk = EarlyRelease
        EarlyReleaseClrOcc = EarlyClrOcc
        AButton = AButtonRef
    End Sub

    Sub UpdateDetails(ByVal ID As String, ByVal SigEntry As String, ByVal SigExit As String, ByVal Conflicting As ArrayList, ByVal NPoints As ArrayList, ByVal RPoints As ArrayList, ByVal Tracks As ArrayList, ByVal MiscTrue As String, ByVal MiscFalse As String, ByVal EarlyRelease As String, ByVal EarlyClrOcc As String, ByVal AButtonRef As String)
        RouteName = ID
        EntrySignal = SigEntry
        ExitSignal = SigExit
        ConflictingRoutes = Conflicting
        PointsNormal = NPoints
        PointsReverse = RPoints
        TrackSections = Tracks
        ILMiscTrue = MiscTrue
        ILMiscFalse = MiscFalse
        EarlyReleaseTk = EarlyRelease
        EarlyReleaseClrOcc = EarlyClrOcc
        AButton = AButtonRef
    End Sub

    'Check Route ID is Present and Unique
    Public Function ValidateID() As Boolean
        Dim result As Boolean = True
        If RouteName = "" Then
            result = False
            MessageBox.Show("Error - Route ID Must be Entered", mrcc)
        Else
            If Data.Routes.ContainsKey(RouteName) Then
                result = False
                MessageBox.Show("Error - Route ID already in Use", mrcc)
            End If
        End If
        Return result
    End Function

    'Check Entry and Exit Combination is unique
    Public Function ValidateNX() As Boolean
        Dim result As Boolean = True
        If EntrySig = ExitSig Then
            result = False
            MessageBox.Show("Error - Entry and Exit Signals must be different", mrcc)
        Else
            If Data.CheckRouteExists(EntrySig, ExitSig) Then
                result = False
                MessageBox.Show("Error - Route from " & EntrySig & " to " & ExitSig & " already exists", mrcc)
            End If
        End If
        Return result
    End Function

    'Checks Route Is Valid
    Public Function ILDataValid() As Boolean
        If ValidateILData() = "" Then
            Return True
        Else
            Return False
        End If
    End Function

    'Checks Route Is Valid and returns Error Report
    Public Function ValidateILData() As String
        Dim i As Integer
        Dim errorreport As String = ""

        'ID is already Validated But Include this one Basic Check
        If RouteName = "" Then errorreport &= "Route ID Not Specified" & vbCrLf
        'Check Entry and Exit Sigs exist and are different
        If EntrySig = "" Then
            errorreport &= "Route " & RouteName & ": No Entry Signal Specified" & vbCrLf
        Else
            If Data.CheckReference(EntrySig) < 0 Or Data.CheckReference(EntrySig) > 8 Then
                errorreport &= "Route " & RouteName & ": Entry Signal " & EntrySig & " is not a Valid Signal Reference" & vbCrLf
            End If
        End If
        If EntrySig = "" Then
            errorreport &= "Route " & RouteName & ": No Exit Signal Specified" & vbCrLf
        Else
            If Data.CheckReference(ExitSig) < 0 Or Data.CheckReference(ExitSig) > 8 Then
                errorreport &= "Route " & RouteName & ": Exit Signal " & ExitSig & " is not a Valid Signal Reference" & vbCrLf
            End If
        End If
        If EntrySig = ExitSig Then
            errorreport &= "Route " & RouteName & ": Entry and Exit Signals must be different" & vbCrLf
        End If

        'Check Misc References are Valid or Blank
        If Not ILMiscTrue = "" Then
            If Not ((Data.CheckReference(ILMiscTrue) = 12) Or (Data.CheckReference(ILMiscTrue) = 14)) Then
                errorreport &= "Route " & RouteName & ": Misc. True Condition " & ILMiscTrue & " is not a Valid Misc. Device" & vbCrLf
            End If
            If AButton = "**AUTO**" Then errorreport &= "Route " & RouteName & ": Auto Routes are not allowed Misc Conditions" & vbCrLf
        End If
        If Not ILMiscFalse = "" Then
            If Not ((Data.CheckReference(ILMiscTrue) = 12) Or (Data.CheckReference(ILMiscTrue) = 14)) Then
                errorreport &= "Route " & RouteName & ": Misc. False Condition " & ILMiscFalse & " is not a Valid Misc. Device" & vbCrLf
            End If
            If AButton = "**AUTO**" Then errorreport &= "Route " & RouteName & ": Auto Routes are not allowed Misc Conditions" & vbCrLf
        End If

        'Check Early Release Track
        If Not EarlyReleaseTk = "" Then
            If Data.CheckReference(EarlyReleaseTk) <> 11 Then
                errorreport &= "Route " & RouteName & ": Early Release Track " & EarlyReleaseTk & " is not a Valid Track Section" & vbCrLf
            End If
            If AButton = "**AUTO**" Then errorreport &= "Route " & RouteName & ": Auto Routes are not allowed Early Release" & vbCrLf
        End If

        'Check A Button Ref
        If Not (AButton = "" Or AButton = "**AUTO**") Then
            If Data.CheckReference(AButton) <> 12 Then
                errorreport &= "Route " & RouteName & ": Auto Button Reference " & AButton & " is not a Valid Auto Button" & vbCrLf
            Else
                If CType(Data.GetByName(AButton), StaticMisc).MiscType <> "Auto" Then
                    errorreport &= "Route " & RouteName & ": Auto Button Reference " & AButton & " is not a Valid Auto Button" & vbCrLf
                End If
            End If
        End If

        'Check Tracks Clear are Valid Tracks
        If TrackSections.Count = 0 Then
            errorreport &= "Route " & RouteName & ": Does not read over any Track Sections" & vbCrLf
        Else
            For i = 1 To TrackSections.Count
                If Not Data.CheckReference(TrackSections(i - 1)) = 11 Then
                    errorreport &= "Route " & RouteName & ": Tracks Clear List, Track " & TrackSections(i - 1) & " is not a Valid Track Section" & vbCrLf
                End If
            Next
        End If
        'Check Conflicting Routes are Valid
        If Conflicting.Count > 0 And AButton = "**AUTO**" Then errorreport &= "Route " & RouteName & ": Auto Routes are not allowed to have Conflicting Routes" & vbCrLf
        For i = 1 To Conflicting.Count
            If Data.CheckRouteExists(Conflicting(i - 1)) = False Then
                errorreport &= "Route " & RouteName & ": Conflicting Route List, Route " & Conflicting(i - 1) & " is not a Valid Route" & vbCrLf
            Else
                'Checks it doesn't conflict with itself!
                If RouteName = Conflicting(i - 1) Then
                    errorreport &= "Route " & RouteName & ": can Not Conflict with itself!" & vbCrLf
                Else
                    'Checks for Corresponding Conflicting Reference to itself in any Conflicting Route
                    If CType(Data.Routes(Conflicting(i - 1)), Route).Conflicting.Contains(RouteName) = False Then
                        errorreport &= "Route " & Conflicting(i - 1) & " does not have Route " & RouteName & " as a Conflicting Route" & vbCrLf
                    End If
                End If
            End If
        Next

        'Check Points References
        'Check Normal Points
        If NPoints.Count > 0 And AButton = "**AUTO**" Then errorreport &= "Route " & RouteName & ": Auto Routes can not read over points" & vbCrLf
        For i = 1 To NPoints.Count
            If Data.CheckReference(NPoints(i - 1)) <> 10 Then
                errorreport &= "Route " & RouteName & ": Points Normal List, Point " & NPoints(i - 1) & " is not a Valid Point" & vbCrLf
            End If
        Next
        'Check Reverse Points
        If RPoints.Count > 0 And AButton = "**AUTO**" Then errorreport &= "Route " & RouteName & ": Auto Routes can not read over points" & vbCrLf
        For i = 1 To RPoints.Count
            If Data.CheckReference(RPoints(i - 1)) <> 10 Then
                errorreport &= "Route " & RouteName & ": Points Reverse List, Point " & RPoints(i - 1) & " is not a Valid Point" & vbCrLf
            Else
                'Check Point is not in both Normal and Reverse List
                If NPoints.Contains(RPoints(i - 1)) = True Then
                    errorreport &= "Route " & RouteName & ": Point " & RPoints(i - 1) & " can Not be called Normal and Reverse by the same route!" & vbCrLf
                Else
                    'Check a Point Pair of Point isn't in Normal List
                    If NPoints.Contains(CType(Data.GetByName(RPoints(i - 1)), Points).PointPair) Then
                        errorreport &= "Route " & RouteName & ": Point " & RPoints(i - 1) & ", Point Pair can Not be called to Opposite Direction by the same route!" & vbCrLf
                    End If
                End If
            End If
        Next

        Return errorreport
    End Function

    Public ReadOnly Property ID() As String
        Get
            Return RouteName
        End Get
    End Property

    Public ReadOnly Property EntrySig() As String
        Get
            Return EntrySignal
        End Get
    End Property

    Public ReadOnly Property ExitSig() As String
        Get
            Return ExitSignal
        End Get
    End Property

    Public ReadOnly Property Conflicting() As ArrayList
        Get
            Return ConflictingRoutes
        End Get
    End Property

    Public ReadOnly Property NPoints() As ArrayList
        Get
            Return PointsNormal
        End Get
    End Property

    Public ReadOnly Property RPoints() As ArrayList
        Get
            Return PointsReverse
        End Get
    End Property

    Public ReadOnly Property TkSections() As ArrayList
        Get
            Return TrackSections
        End Get
    End Property

    Public ReadOnly Property MiscTrue() As String
        Get
            Return ILMiscTrue
        End Get
    End Property

    Public ReadOnly Property MiscFalse() As String
        Get
            Return ILMiscFalse
        End Get
    End Property

    Public ReadOnly Property EarlyReleaseTrack() As String
        Get
            Return EarlyReleaseTk
        End Get
    End Property

    Public ReadOnly Property EarlyClrOcc() As String
        Get
            Return EarlyReleaseClrOcc
        End Get
    End Property

    Public Property AButtonRef() As String
        Get
            Return AButton
        End Get
        Set(ByVal Value As String)
            'Can Only be Set as part of Auto Signal functionality
            If Not (Value = "**AUTO**" Or Value = "") Then
                MessageBox.Show("Invalid Attempt to Set AButtonRef for Route " & Me.ID, mrcc)
            Else
                If AButtonRef <> Value Then
                    'Change is Needed
                    If Not (AButtonRef = "" Or AButtonRef = "**AUTO**") Then
                        'AButtonRef set to an A Button
                        If Value = "**AUTO**" Then
                            'Warn before overwriting AButton Ref with AUTO
                            MessageBox.Show("Warning: - " & vbCrLf & "Route " & Me.ID & " had Auto Button " & Me.AButtonRef & " referenced." & vbCrLf & "This has been overwritten and Route designated Auto.", mrcc)
                            AButton = Value
                        End If
                        'Don't Overwrite AButton Ref with ""
                    Else
                        'Set AUTO or "" as required - No existing A button Ref
                        AButton = Value
                    End If
                End If
            End If
        End Set
    End Property

    Public Property USet() As Boolean
        Get
            Return Me.RouteSet
        End Get
        Set(ByVal Value As Boolean)
            Me.RouteSet = Value
        End Set
    End Property

    Public Property AButtonSet() As Boolean
        Get
            Return Me.APressed
        End Get
        Set(ByVal Value As Boolean)
            Me.APressed = Value
        End Set
    End Property

    Public Property TracksClear() As Boolean
        Get
            Return Me.TracksClr
        End Get
        Set(ByVal Value As Boolean)
            Me.TracksClr = Value
        End Set
    End Property

    Public Sub SetAutoRoutes()
        If Me.AButtonRef = "**AUTO**" Then
            Me.AButtonSet = True
            Me.CallRoute()
        End If
    End Sub

    Public Function CallRoute() As String
        Dim FeedbackText As String = ""
        If Me.USet = False Then
            Dim i As Integer
            Dim errorreport As String = ""

            'Check Conflicting Routes
            For i = 0 To Conflicting.Count - 1
                If Data.GetRoute(Conflicting(i)).USet = True Then
                    errorreport &= "Conflicting Route, " & Conflicting(i) & " is Set" & vbCrLf
                End If
            Next

            'Check Tracks Clear
            If Me.TracksClear = False Then
                errorreport &= "Some Tracks within the Route are Occupied" & vbCrLf
            End If

            'Check Misc Conditions
            If Me.ILMiscTrue <> "" Then
                If Data.GetMiscState(Me.ILMiscTrue) = False Then
                    errorreport &= "Misc. Condition, " & Me.ILMiscTrue & " is False" & vbCrLf
                End If
            End If
            If Me.ILMiscFalse <> "" Then
                If Data.GetMiscState(Me.ILMiscFalse) = True Then
                    errorreport &= "Misc. Condition, " & Me.ILMiscFalse & " is True" & vbCrLf
                End If
            End If
            If errorreport = "" Then
                'Call Points
                For i = 0 To NPoints.Count - 1
                    errorreport &= Data.GetPoint(NPoints(i)).CallNormal("Route")
                Next
                For i = 0 To RPoints.Count - 1
                    errorreport &= Data.GetPoint(RPoints(i)).CallReverse("Route")
                Next
            End If
            If errorreport = "" Then
                'Set Route
                For i = 0 To TkSections.Count - 1
                    Data.GetTKSection(TkSections(i)).USet = True
                    If Me.AButtonSet Then
                        Data.GetTKSection(TkSections(i)).APressed = True
                    End If
                Next
                Me.USet = True
            Else
                'Route Failed to Set Message
                If ModeChangeFlag = False Then
                    FeedbackText = "Route " & RouteName & " failed to set : -" & vbCrLf & vbCrLf & errorreport
                Else
                    FeedbackText = "Route " & RouteName & " failed to set : -" & vbCrLf & vbCrLf & errorreport & vbCrLf & vbCrLf & "PLEASE SET THIS ROUTE MANUALLY WHEN IT IS AVAILABLE"
                End If
            End If
        End If
        Return FeedbackText
    End Function

    Public Sub CancelRoute()
        If Not (ModeChangeFlag = False And AButton = "**AUTO**") Then
            Dim i As Integer

            'Pull A Button
            If Not (AButton = "" Or AButton = "**AUTO**") Then
                Data.GetStaticMisc(AButton).CancelMisc()
            Else
                If AButton = "**AUTO**" Then Me.APressed = False
            End If

            For i = 0 To TrackSections.Count - 1
                Data.GetTKSection(TrackSections(i)).USet = False
                Data.GetTKSection(TrackSections(i)).TORRAv = False

                Data.GetTKSection(TrackSections(i)).APressed = False
            Next

            USet = False
            APressed = False
        End If
    End Sub

    Public Sub Update()

        Dim i As Integer
        'Poll A Button
        If Not (AButton = "" Or AButton = "**AUTO**") Then
            If USet = True Then
                APressed = Data.GetMiscState(AButton)
                For i = 0 To TrackSections.Count - 1
                    Data.GetTKSection(TrackSections(i)).APressed = APressed
                Next
                'Else
                '    If APressed = False And Data.GetMiscState(AButton) = True Then
                '        Data.GetStaticMisc(AButton).CancelMisc()
                '        MessageBox.Show("Unable to Set " & AButton & " : Route must be set first", mrcc)
                '    End If
            End If
        End If

        'Check if Tracks are Clear
        'Check if Tracks have no Route Set and Release Route if Required
        Dim TempTkOcc As Boolean = True
        Dim TempUSet As Boolean = False
        For i = 0 To TrackSections.Count - 1
            If Data.GetTKSection(TrackSections(i)).TOcc = True Then TempTkOcc = False
            If Data.GetTKSection(TrackSections(i)).USet = True Then TempUSet = True
        Next
        Me.TracksClear = TempTkOcc
        If TempUSet = False Then
            For i = 0 To TrackSections.Count - 1
                Data.GetTKSection(TrackSections(i)).TORRAv = False
            Next i
            Me.USet = False
        End If

        'Check for Early Release
        If EarlyReleaseTk <> "" Then
            If (EarlyReleaseClrOcc = "Clr" And ERTkOcc = True And Data.GetTKSection(EarlyReleaseTk).TOcc = False) Or (EarlyReleaseClrOcc = "Occ" And ERTkOcc = False And Data.GetTKSection(EarlyReleaseTk).TOcc = True) Then
                'Check Route Set and A Button Not pressed
                If USet = True And APressed = False Then
                    'Early Release
                    For i = 0 To TrackSections.Count - 1
                        Data.GetTKSection(TrackSections(i)).USet = False
                        Data.GetTKSection(TrackSections(i)).TORRAv = False
                    Next i
                    Me.USet = False
                End If
            End If

            'Update ERTkOcc for next cycle
            ERTkOcc = Data.GetTKSection(EarlyReleaseTk).TOcc
        End If

    End Sub

End Class

'POINT KEYS
'**********
<Serializable()> Public Class PointKey
    Inherits ScreenElements

    Private AssociatedPointID As String = ""
    Private SymbolCode As Integer = 158
    Public Sub New(ByVal x As Integer, ByVal y As Integer, ByVal PointID As String)
        MyBase.New(x, y)
        AssociatedPointID = PointID
    End Sub

    Public Sub UpdateDetails(ByVal PointID As String)
        AssociatedPointID = PointID
    End Sub

    Public ReadOnly Property Symbol() As Integer
        Get
            Return SymbolCode
        End Get
    End Property

    Public ReadOnly Property PointID() As String
        Get
            Return AssociatedPointID
        End Get
    End Property

    Public Sub UpdateKey(ByVal NCR As String)
        Select Case (NCR)
            Case "N"
                If SymbolCode <> 158 Then
                    SymbolCode = 158
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            Case "C"
                If SymbolCode <> 156 Then
                    SymbolCode = 156
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
            Case "R"
                If SymbolCode <> 159 Then
                    SymbolCode = 159
                    GUI.UpdateDisplay(MyBase.X_Pos, MyBase.Y_Pos)
                End If
        End Select
    End Sub

    Public Function Validate() As Boolean
        If Data.CheckReference(AssociatedPointID) = 10 Then
            Return True
        Else
            MessageBox.Show(AssociatedPointID & " is Not a Valid Point", mrcc)
            Return False
        End If
    End Function

    Public Sub Update()
        UpdateKey(Data.GetPoint(AssociatedPointID).Key)
    End Sub

    Public Sub EnterDesignMode()
        UpdateKey("N")
    End Sub
End Class